"""
It keeps the service configuration parameters like maximum running threads, number of processes, etc. ,
which can be configured in CS.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC.Core.Utilities import Network, List
from DIRAC.ConfigurationSystem.Client.ConfigurationData import gConfigurationData
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.DISET.private.Protocols import gDefaultProtocol


class ServiceConfiguration:
    def __init__(self, nameList):
        self.serviceName = nameList[0]
        self.serviceURL = None
        self.nameList = nameList
        self.pathList = []
        for svcName in nameList:
            self.pathList.append(PathFinder.getServiceSection(svcName))

    def getOption(self, optionName):
        if optionName[0] == "/":
            return gConfigurationData.extractOptionFromCFG(optionName)
        for path in self.pathList:
            value = gConfigurationData.extractOptionFromCFG("%s/%s" % (path, optionName))
            if value:
                return value
        return None

    def getAddress(self):
        return ("", self.getPort())

    def getHandlerLocation(self):
        return self.getOption("HandlerPath")

    def getName(self):
        return self.serviceName

    def setURL(self, sURL):
        self.serviceURL = sURL

    def __getCSURL(self, URL=None):
        optionValue = self.getOption("URL")
        if optionValue:
            return optionValue
        return URL

    def registerAlsoAs(self):
        optionValue = self.getOption("RegisterAlsoAs")
        if optionValue:
            return List.fromChar(optionValue)
        else:
            return []

    def getMaxThreads(self):
        try:
            return int(self.getOption("MaxThreads"))
        except Exception:
            return 15

    def getMinThreads(self):
        try:
            return int(self.getOption("MinThreads"))
        except Exception:
            return 1

    def getMaxWaitingPetitions(self):
        try:
            return int(self.getOption("MaxWaitingPetitions"))
        except Exception:
            return 100

    def getMaxMessagingConnections(self):
        try:
            return int(self.getOption("MaxMessagingConnections"))
        except Exception:
            return 20

    def getMaxThreadsForMethod(self, actionType, method):
        try:
            return int(self.getOption("ThreadLimit/%s/%s" % (actionType, method)))
        except Exception:
            return 15

    def getCloneProcesses(self):
        try:
            return int(self.getOption("CloneProcesses"))
        except Exception:
            return 0

    def getPort(self):
        try:
            return int(self.getOption("Port"))
        except Exception:
            return 9876

    def getProtocol(self):
        optionValue = self.getOption("Protocol")
        if optionValue:
            return optionValue
        return gDefaultProtocol

    def getHostname(self):
        hostname = self.getOption("/DIRAC/Hostname")
        if not hostname:
            return Network.getFQDN()
        return hostname

    def getURL(self):
        """
        Build the service URL
        """
        if self.serviceURL:
            return self.serviceURL
        protocol = self.getProtocol()
        serviceURL = self.__getCSURL()
        if serviceURL:
            if serviceURL.find(protocol) != 0:
                urlFields = serviceURL.split(":")
                urlFields[0] = protocol
                serviceURL = ":".join(urlFields)
                self.setURL(serviceURL)
            return serviceURL
        hostName = self.getHostname()
        port = self.getPort()
        serviceURL = "%s://%s:%s/%s" % (protocol, hostName, port, self.getName())
        if serviceURL[-1] == "/":
            serviceURL = serviceURL[:-1]
        self.setURL(serviceURL)
        return serviceURL

    def getContextLifeTime(self):
        optionValue = self.getOption("ContextLifeTime")
        try:
            return int(optionValue)
        except Exception:
            return 21600



"""
   CloudEndpoint is a base class for the clients used to connect to different
   cloud providers
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import os

from DIRAC import S_ERROR, S_OK
from DIRAC.Resources.Cloud.Utilities import createUserDataScript, createPilotDataScript, createCloudInitScript

__RCSID__ = "$Id$"


class Endpoint(object):
    """Endpoint base class"""

    def __init__(self, parameters=None, bootstrapParameters=None):
        # logger
        if not parameters:
            parameters = {}
        self.parameters = parameters
        if not bootstrapParameters:
            bootstrapParameters = {}
        self.bootstrapParameters = bootstrapParameters
        self.valid = False
        self.proxy = None

    def isValid(self):
        return self.valid

    def setParameters(self, parameters):
        self.parameters.update(parameters)

    def setBootstrapParameters(self, bootstrapParameters):
        self.bootstrapParameters.update(bootstrapParameters)

    def getParameterDict(self):
        return self.parameters

    def setProxy(self, proxy):
        self.proxy = proxy

    def getProxyFileLocation(self):
        if not self.proxy:
            self.proxy = self.parameters.get("Proxy", os.environ.get("X509_USER_PROXY"))
            if not self.proxy:
                return S_ERROR("Can not find proxy")
        return S_OK(self.proxy)

    def _createUserDataScript(self):

        bootType = self.bootstrapParameters.get("BootType", "pilot")
        if bootType.lower() == "pilot":
            return createPilotDataScript(self.parameters, self.bootstrapParameters)
        elif bootType.lower() == "user":
            return createUserDataScript(self.parameters)
        elif bootType.lower() == "cloudinit":
            return createCloudInitScript(self.parameters, self.bootstrapParameters)



""" Module that contains client access to the Pilots handler.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC.Core.Base.Client import Client, createClient


@createClient("WorkloadManagement/PilotManager")
class PilotManagerClient(Client):
    """PilotManagerClient sets url for the PilotManagerHandler."""

    def __init__(self, url=None, **kwargs):
        """
        Sets URL for PilotManager handler

        :param self: self reference
        :param url: url of the PilotManagerHandler
        :param kwargs: forwarded to the Base Client class
        """

        super(PilotManagerClient, self).__init__(**kwargs)

        if not url:
            self.serverURL = "WorkloadManagement/PilotManager"

        else:
            self.serverURL = url



""" Module that holds the DataStore Client class
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import time
import random
import copy
import threading

from DIRAC import S_OK, S_ERROR, gLogger, gConfig
from DIRAC.Core.Base.Client import Client, createClient
from DIRAC.Core.Utilities import DEncode
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.Client.Operation import Operation
from DIRAC.RequestManagementSystem.Client.ReqClient import ReqClient
from DIRAC.Core.Utilities.DErrno import ERMSUKN


@createClient("Accounting/DataStore")
class DataStoreClient(Client):
    """
    Class providing front end access to DIRAC Accounting DataStore Service
     - It allows to reduce the interactions with the server by building and list of
    pending Registers to be sent that are sent in a bundle using the commit method.
     - In case the DataStore is down Registers are sent as DISET requests.
    """

    def __init__(self, retryGraceTime=0, **kwargs):
        """Simple constructor

        :params int retryGraceTime: the seconds to wait before sending records to the failover
        """
        super(DataStoreClient, self).__init__(**kwargs)
        self.setServer("Accounting/DataStore")

        self.__maxRecordsInABundle = 5000
        self.__registersList = []
        self.__maxTimeRetrying = retryGraceTime
        self.__lastSuccessfulCommit = time.time()
        self.__failoverEnabled = not gConfig.getValue("/LocalSite/DisableFailover", False)
        self.__registersListLock = threading.RLock()
        self.__commitTimer = threading.Timer(5, self.commit)

    def __checkBaseType(self, obj):
        """Check to find that the class inherits from the Base Type"""
        for parent in obj.__bases__:
            if parent.__name__ == "BaseAccountingType":
                return True
            if self.__checkBaseType(parent):
                return True
        return False

    def addRegister(self, register):
        """Add a register to the list to be sent

        :params BaseAccountingType register: the accounting record to send.
        """
        if not self.__checkBaseType(register.__class__):
            return S_ERROR("register is not a valid type (has to inherit from BaseAccountingType")
        retVal = register.checkValues()
        if not retVal["OK"]:
            return retVal
        if gConfig.getValue("/LocalSite/DisableAccounting", False):
            return S_OK()

        self.__registersList.append(copy.deepcopy(list(register.getValues())))

        return S_OK()

    def disableFailover(self):
        self.__failoverEnabled = False

    def commit(self):
        """
        Send the registers in a bundle mode
        """
        rpcClient = self._getRPC()
        sent = 0

        # create a local reference and prevent other running commits
        # to take the same data second time
        self.__registersListLock.acquire()
        registersList = self.__registersList
        self.__registersList = []
        self.__registersListLock.release()

        try:
            while registersList:
                registersToSend = registersList[: self.__maxRecordsInABundle]
                retVal = rpcClient.commitRegisters(registersToSend)
                if retVal["OK"]:
                    self.__lastSuccessfulCommit = time.time()
                else:
                    if self.__failoverEnabled and time.time() - self.__lastSuccessfulCommit > self.__maxTimeRetrying:
                        gLogger.warn("Error sending accounting record", retVal["Message"])
                        gLogger.verbose("Sending accounting records to failover")
                        result = _sendToFailover(retVal["rpcStub"])
                        if not result["OK"]:
                            return result
                        gLogger.debug("Sent accounting record to failover")
                    else:
                        gLogger.error("Error sending accounting record", retVal["Message"])
                        return S_ERROR("Cannot commit data to DataStore service")
                sent += len(registersToSend)
                del registersList[: self.__maxRecordsInABundle]
        except Exception as e:  # pylint: disable=broad-except
            gLogger.exception("Error committing", lException=e)
            return S_ERROR("Error committing %s" % repr(e).replace(",)", ")"))
        finally:
            # if something is left because of an error return it to the main list
            self.__registersList.extend(registersList)

        return S_OK(sent)

    def delayedCommit(self):
        """
        If needed start a timer that will run the commit later
        allowing to send more registers at once (reduces overheads).
        """

        if not self.__commitTimer.is_alive():
            self.__commitTimer = threading.Timer(5, self.commit)
            self.__commitTimer.start()

        return S_OK()

    def remove(self, register):
        """
        Remove a Register from the Accounting DataStore
        """
        if not self.__checkBaseType(register.__class__):
            return S_ERROR("register is not a valid type (has to inherit from BaseAccountingType")
        retVal = register.checkValues()
        if not retVal["OK"]:
            return retVal
        if gConfig.getValue("/LocalSite/DisableAccounting", False):
            return S_OK()
        return self._getRPC().remove(*register.getValues())


def _sendToFailover(rpcStub):
    """Create a ForwardDISET operation for failover"""
    try:
        request = Request()
        request.RequestName = "Accounting.DataStore.%s.%s" % (time.time(), random.random())
        forwardDISETOp = Operation()
        forwardDISETOp.Type = "ForwardDISET"
        forwardDISETOp.Arguments = DEncode.encode(rpcStub)
        request.addOperation(forwardDISETOp)

        return ReqClient().putRequest(request)

    # We catch all the exceptions, because it should never crash
    except Exception as e:  # pylint: disable=broad-except
        return S_ERROR(ERMSUKN, "Exception sending accounting failover request: %s" % repr(e))


gDataStoreClient = DataStoreClient()



""" Collection of DIRAC useful adler32 related tools.
    By default on Error they return None.

   .. warning::

      On error False is returned.

   .. warning::

      All exceptions report to the stdout.
"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import six
from zlib import adler32


def intAdlerToHex(intAdler):
    """Change adler32 checksum base from decimal to hex.

    :param integer intAdler: adler32 checksum
    :return: 8 digit hex string
    """
    try:
        # Will always be 8 hex digits made from a positive integer
        return hex(intAdler & 0xFFFFFFFF).lower().replace("l", "").replace("x", "0000")[-8:]
    except Exception as error:
        print(repr(error).replace(",)", ")"))
        return False


def hexAdlerToInt(hexAdler, pos=True):
    """Change hex base to decimal for adler32 checksum.

    :param mixed hexAdler: hex based adler32 checksum integer or a string
    :param boolean pos: flag to determine sign (default True = positive)
    """
    if isinstance(hexAdler, six.integer_types):
        return hexAdler & 0xFFFFFFFF
    # First make sure we can parse the hex properly
    if hexAdler == "False" or hexAdler == "-False":
        return False
    hexAdler = hexAdler.lower().replace("l", "")
    hexAdler = hexAdler[-8:]
    hexAdler = hexAdler.replace("x", "0")
    if not pos:
        hexAdler = "-%s" % hexAdler
    try:
        # Will always try to return the positive integer value of the provided hex
        return int(hexAdler, 16) & 0xFFFFFFFF
    except Exception as error:
        print(repr(error).replace(",)", ")"))
        return False


def compareAdler(adler1, adler2):
    """Check equality between two adler32 checksums.

    :param adler1: 1st checksum
    :param adler2: 2nd checksum
    :return: True (False) if cheksums are (not) equal
    """
    adler1s = [hexAdlerToInt(adler1), hexAdlerToInt(adler1, False)]
    if False in adler1s:
        adler1s.remove(False)
    adler2s = [hexAdlerToInt(adler2), hexAdlerToInt(adler2, False)]
    if False in adler2s:
        adler2s.remove(False)
    for adler1 in adler1s:
        if adler1 in adler2s:
            return True
    return False


def fileAdler(fileName):
    """Calculate alder32 checksum of the supplied file.

    :param str fileName: path to file
    """

    def readChunk(fd, size=1048576):
        """Return data from file descriptor in chunk of size size.

        :param fd: file descriptor
        :param integer size: size of data chunk in bytes (default 1024 * 1024 = 1048576)
        """
        while True:
            data = fd.read(size)
            if not data:
                break
            yield data

    try:
        with open(fileName, "rb") as inputFile:
            myAdler = 1
            for data in readChunk(inputFile):
                myAdler = adler32(data, myAdler)
            return intAdlerToHex(myAdler)
    except Exception as error:
        print(repr(error).replace(",)", ")"))
        return False


def stringAdler(string):
    """Calculate adler32 of the supplied string.

    :param str string: data
    """
    try:
        intAdler = adler32(string.encode())
        return intAdlerToHex(intAdler)
    except Exception as error:
        print(repr(error).replace(",)", ")"))
        return False


if __name__ == "__main__":
    import sys

    for p in sys.argv[1:]:
        print("%s : %s" % (p, fileAdler(p)))



"""

This module module is used to generate the CAs and CRLs (revoked certificates)

Example::

  from DIRAC.Core.Security import Utilities

  retVal = Utilities.generateRevokedCertsFile()
  if retVal['OK']:
    cl = Elasticsearch( self.__url,
                        timeout = self.__timeout,
                        use_ssl = True,
                        verify_certs = True,
                        ca_certs = retVal['Value'] )

or::

  retVal = Utilities.generateCAFile('/WebApp/HTTPS/Cert')
  if retVal['OK']:
    sslops = dict( certfile = CertificateMgmt.getCert(/WebApp/HTTPS/Cert),
                   keyfile = CertificateMgmt.getCert(/WebApp/HTTPS/Key),
                   cert_reqs = ssl.CERT_OPTIONAL,
                   ca_certs = retVal['Value'],
                   ssl_version = ssl.PROTOCOL_TLSv1 )

  srv = tornado.httpserver.HTTPServer( self.__app, ssl_options = sslops, xheaders = True )

Note: If you wan to make sure that the CA is up to date, better to use the BundleDeliveryClient.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function
import os
import tempfile

from DIRAC.Core.Security import X509Chain, X509CRL
from DIRAC.Core.Security import Locations
from DIRAC import gLogger, S_OK, S_ERROR


def generateCAFile(location=None):
    """

    Generate/find a single CA file with all the PEMs

    :param str location: we can specify a specific CS location
                         where it's written a directory where to find the CAs and CRLs
    :return: directory where the file cas.pem which contains all certificates is found/created

    """
    caDir = Locations.getCAsLocation()
    if not caDir:
        return S_ERROR("No CAs dir found")

    # look in what's normally /etc/grid-security/certificates
    if os.path.isfile(os.path.join(os.path.dirname(caDir), "cas.pem")):
        return S_OK(os.path.join(os.path.dirname(caDir), "cas.pem"))

    # look in what's normally /opt/dirac/etc/grid-security
    diracCADirPEM = os.path.join(os.path.dirname(Locations.getHostCertificateAndKeyLocation(location)[0]), "cas.pem")
    if os.path.isfile(diracCADirPEM):
        return S_OK(diracCADirPEM)

    # Now we create it in tmpdir
    fn = tempfile.mkstemp(prefix="cas.", suffix=".pem")[1]
    try:
        with open(fn, "wb") as fd:
            for caFile in os.listdir(caDir):
                caFile = os.path.join(caDir, caFile)
                chain = X509Chain.X509Chain()
                result = chain.loadChainFromFile(caFile)
                if not result["OK"]:
                    continue

                expired = chain.hasExpired()
                if not expired["OK"] or expired["Value"]:
                    continue
                fd.write(chain.dumpAllToString()["Value"])

        gLogger.info("CAs used from: %s" % str(fn))
        return S_OK(fn)
    except IOError as err:
        gLogger.warn(err)

    return S_ERROR("Could not find/generate CAs")


def generateRevokedCertsFile(location=None):
    """

    Generate a single CA file with all the PEMs

    :param str location: we can specify a specific CS location
                         where it's written a directory where to find the CAs and CRLs
    :return: directory where the file crls.pem which contains all CRLs is created

    """
    caDir = Locations.getCAsLocation()
    if not caDir:
        return S_ERROR("No CAs dir found")

    # look in what's normally /etc/grid-security/certificates
    if os.path.isfile(os.path.join(os.path.dirname(caDir), "crls.pem")):
        return S_OK(os.path.join(os.path.dirname(caDir), "crls.pem"))

    # look in what's normally /opt/dirac/etc/grid-security
    diracCADirPEM = os.path.join(os.path.dirname(Locations.getHostCertificateAndKeyLocation(location)[0]), "crls.pem")
    if os.path.isfile(diracCADirPEM):
        return S_OK(diracCADirPEM)

    # Now we create it in tmpdir
    fn = tempfile.mkstemp(prefix="crls", suffix=".pem")[1]
    try:
        with open(fn, "wb") as fd:
            for caFile in os.listdir(caDir):
                caFile = os.path.join(caDir, caFile)
                result = X509CRL.X509CRL.instanceFromFile(caFile)
                if not result["OK"]:
                    continue
                chain = result["Value"]
                fd.write(chain.dumpAllToString()["Value"])
            return S_OK(fn)
    except IOError as err:
        gLogger.warn(err)

    return S_ERROR("Could not find/generate CRLs")



#!/usr/bin/env python
"""
Get status of the available Storage Elements

Usage:
  dirac-dms-show-se-status [<options>]

Example:
  $ dirac-dms-show-se-status
  Storage Element               Read Status    Write Status
  DIRAC-USER                         Active          Active
  IN2P3-disk                         Active          Active
  IPSL-IPGP-disk                     Active          Active
  IRES-disk                        InActive        InActive
  M3PEC-disk                         Active          Active
  ProductionSandboxSE                Active          Active
"""
from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

from DIRAC import S_OK, exit as DIRACexit
from DIRAC.Core.Utilities.DIRACScript import DIRACScript as Script

__RCSID__ = "$Id$"

vo = None


def setVO(arg):
    global vo
    vo = arg
    return S_OK()


allVOsFlag = False


def setAllVO(arg):
    global allVOsFlag
    allVOsFlag = True
    return S_OK()


noVOFlag = False


def setNoVO(arg):
    global noVOFlag, allVOsFlag
    noVOFlag = True
    allVOsFlag = False
    return S_OK()


@Script()
def main():
    global vo
    global noVOFlag
    global allVOsFlag

    Script.registerSwitch("V:", "vo=", "Virtual Organization", setVO)
    Script.registerSwitch("a", "all", "All Virtual Organizations flag", setAllVO)
    Script.registerSwitch("n", "noVO", "No Virtual Organizations assigned flag", setNoVO)

    Script.parseCommandLine()

    from DIRAC import gConfig, gLogger
    from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus
    from DIRAC.Core.Utilities.PrettyPrint import printTable
    from DIRAC.Core.Security.ProxyInfo import getVOfromProxyGroup

    storageCFGBase = "/Resources/StorageElements"

    res = gConfig.getSections(storageCFGBase, True)
    if not res["OK"]:
        gLogger.error("Failed to get storage element info")
        gLogger.error(res["Message"])
        DIRACexit(1)

    gLogger.info("%s %s %s" % ("Storage Element".ljust(25), "Read Status".rjust(15), "Write Status".rjust(15)))

    seList = sorted(res["Value"])

    resourceStatus = ResourceStatus()

    res = resourceStatus.getElementStatus(seList, "StorageElement")
    if not res["OK"]:
        gLogger.error("Failed to get StorageElement status for %s" % str(seList))
        DIRACexit(1)

    fields = ["SE", "ReadAccess", "WriteAccess", "RemoveAccess", "CheckAccess"]
    records = []

    if vo is None and not allVOsFlag:
        result = getVOfromProxyGroup()
        if not result["OK"]:
            gLogger.error("Failed to determine the user VO")
            DIRACexit(1)
        vo = result["Value"]

    for se, statusDict in res["Value"].items():

        # Check if the SE is allowed for the user VO
        if not allVOsFlag:
            voList = gConfig.getValue("/Resources/StorageElements/%s/VO" % se, [])
            if noVOFlag and voList:
                continue
            if voList and vo not in voList:
                continue

        record = [se]
        for status in fields[1:]:
            value = statusDict.get(status, "Unknown")
            record.append(value)
        records.append(record)

    printTable(fields, records, numbering=False, sortField="SE")

    DIRACexit(0)


if __name__ == "__main__":
    main()



"""
:mod: Utils

Module that collects utility functions.

"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

import fnmatch

from DIRAC import gConfig, S_OK
from DIRAC.Core.Utilities import List
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations


def voimport(base_mod):
    """
    Function to import from extensions, if not found, tries from DIRAC.
    """

    for ext in gConfig.getValue("DIRAC/Extensions", []):

        try:
            return __import__(ext + base_mod, globals(), locals(), ["*"])
        except ImportError:
            continue
    # If not found in extensions, import it in DIRAC base.
    return __import__(base_mod, globals(), locals(), ["*"])


def getCSTree(csPath=""):
    """
    Gives the configuration rooted at path in a Python dict. The
    result is a Python dictionary that reflects the structure of the
    configuration file.
    """

    opHelper = Operations()

    def getCSTreeAsDict(treePath):
        """
        Function to recursively iterate over a CS tree
        """

        csTreeDict = {}

        opts = opHelper.getOptionsDict(treePath)
        if opts["OK"]:

            opts = opts["Value"]

            for optKey, optValue in opts.items():
                if optValue.find(",") > -1:
                    optValue = List.fromChar(optValue)
                else:
                    optValue = [optValue]
                csTreeDict[optKey] = optValue

        secs = opHelper.getSections(treePath)
        if secs["OK"]:

            secs = secs["Value"]

            for sec in secs:

                secTree = getCSTreeAsDict("%s/%s" % (treePath, sec))
                if not secTree["OK"]:
                    return secTree

                csTreeDict[sec] = secTree["Value"]

        return S_OK(csTreeDict)

    return getCSTreeAsDict(csPath)


def configMatch(candidateParams, configParams):
    """
    For a given configuration, the candidate will be rejected if:

      - it is missing at least one of the params in the config
      - if a param of the candidate does not match the config params
      - if a candidate param is None, is considered as wildcard

    """

    for key in candidateParams:

        if key not in configParams:
            # The candidateParams is missing one of the parameters required
            # return False
            continue

        if candidateParams[key] is None:
            # None is assumed to be a wildcard (*)
            continue

        cParameter = candidateParams[key]
        if not isinstance(cParameter, list):
            cParameter = [cParameter]

        # We allow using UNIX-like regular expression ( wild-cards ) on the CS
        _matches = False
        for configItem in configParams[key]:
            if fnmatch.filter(set(cParameter), configItem):
                _matches = True
                break

        if not _matches:
            return False

    return True


################################################################################
# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF



"""Test ComponentSupervisionAgent."""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import unittest
import sys
from datetime import datetime, timedelta
from mock import MagicMock, call, patch
import psutil

import DIRAC
from DIRAC import gLogger
from DIRAC import S_OK, S_ERROR

import DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent as MAA
from DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent import ComponentSupervisionAgent, NO_RESTART


__RCSID__ = "$Id$"


def clientMock(ret):
    """Return an Client which returns **ret**."""
    clientModuleMock = MagicMock(name="Client Module")
    clientClassMock = MagicMock(name="Client Class")
    clientClassMock.ping.return_value = ret
    clientModuleMock.return_value = clientClassMock
    return clientModuleMock


def mockComponentSection(*_args, **kwargs):
    """Mock the PathFinder.getComponentSection to return individual componentSections."""
    system = kwargs.get("system")
    component = kwargs.get("component")
    return "/Systems/%s/Production/Services/%s" % (system, component)


def mockURLSection(*_args, **kwargs):
    """Mock the PathFinder.getSystemURLSection to return individual componentSections."""
    system = kwargs.get("system")
    return "/Systems/%s/Production/URLs/" % system


class TestComponentSupervisionAgent(unittest.TestCase):
    """TestComponentSupervisionAgent class."""

    def setUp(self):
        """Set up test environment."""
        self.agent = MAA
        self.agent.AgentModule = MagicMock()
        self.agent.NotificationClient = MagicMock(
            spec=DIRAC.FrameworkSystem.Client.NotificationClient.NotificationClient
        )

        self.restartAgent = ComponentSupervisionAgent()
        self.restartAgent.log = gLogger
        self.restartAgent.log.setLevel("DEBUG")
        self.restartAgent.sysAdminClient = MagicMock()
        self.restartAgent.csAPI = MagicMock()
        self.restartAgent.enabled = True
        self.restartAgent.restartAgents = True
        self.restartAgent.doNotRestartInstancePattern = ["Foo"]

    @classmethod
    def tearDownClass(cls):
        """Remove monitoragent module after tests to avoid side effects."""
        sys.modules.pop("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent")

    @staticmethod
    def getPSMock():
        """Mock psutil."""
        psMock = MagicMock(name="psutil")
        procMock2 = MagicMock(name="process2kill")
        psMock.wait_procs.return_value = ("gone", [procMock2])
        procMock = MagicMock(name="process")
        procMock.children.return_value = []
        psMock.Process.return_value = procMock
        return psMock

    def test_init(self):
        """Test the init function."""
        self.assertIsInstance(self.restartAgent, ComponentSupervisionAgent)
        self.assertIsInstance(self.restartAgent.nClient, MagicMock)
        self.assertIsInstance(self.restartAgent.sysAdminClient, MagicMock)
        self.assertTrue(self.restartAgent.enabled)
        self.assertEqual(self.restartAgent.addressFrom, "")

    def test_begin_execution(self):
        """Test for the beginExecution function."""
        self.restartAgent.accounting["Junk"]["Funk"] = 1
        self.restartAgent.am_getOption = MagicMock()
        getOptionCalls = [
            call("Setup", self.restartAgent.setup),
            call("EnableFlag", True),
            call("MailTo", self.restartAgent.addressTo),
            call("MailFrom", self.restartAgent.addressFrom),
        ]

        self.restartAgent.getRunningInstances = MagicMock(return_value=S_OK())
        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.CSAPI", new=MagicMock()):
            self.restartAgent.beginExecution()

        self.restartAgent.am_getOption.assert_has_calls(getOptionCalls, any_order=True)
        self.restartAgent.getRunningInstances.assert_called()

        # accounting dictionary should be cleared
        self.assertEqual(self.restartAgent.accounting, {})

    def test_send_notification(self):
        """Test for the sendNotification function."""
        self.restartAgent.addressTo = ["foo@bar.baz"]
        self.restartAgent.errors = []
        self.restartAgent.accounting = {}

        # send mail should not be called if there are no errors and accounting information
        self.restartAgent.sendNotification()
        self.restartAgent.nClient.sendMail.assert_not_called()

        # send mail should be called if there are errors but no accounting information
        self.restartAgent.errors = ["some error"]
        self.restartAgent.sendNotification()
        self.restartAgent.nClient.sendMail.assert_called()

        # send email should be called if there is accounting information but no errors
        self.restartAgent.nClient.sendMail.reset_mock()
        self.restartAgent.errors = []
        self.restartAgent.accounting = {"Agent1": {"LogAge": 123, "Treatment": "Agent Restarted"}}
        self.restartAgent.sendNotification()
        self.restartAgent.nClient.sendMail.assert_called()

        # try sending email to all addresses even if we get error for sending email to some address
        self.restartAgent.nClient.sendMail.reset_mock()
        self.restartAgent.errors = ["some error"]
        self.restartAgent.addressTo = ["name1@cern.ch", "name2@cern.ch"]
        self.restartAgent.nClient.sendMail.return_value = S_ERROR()
        self.restartAgent.sendNotification()
        self.assertEqual(len(self.restartAgent.nClient.sendMail.mock_calls), len(self.restartAgent.addressTo))

        # accounting dict and errors list should be cleared after notification is sent
        self.assertEqual(self.restartAgent.accounting, {})
        self.assertEqual(self.restartAgent.errors, [])

    @patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.gConfig", new=MagicMock())
    @patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.PathFinder.getComponentSection", new=MagicMock())
    def test_get_running_instances(self):
        """Test for the getRunningInstances function."""
        self.restartAgent.sysAdminClient.getOverallStatus = MagicMock()
        self.restartAgent.sysAdminClient.getOverallStatus.return_value = S_ERROR()

        res = self.restartAgent.getRunningInstances(instanceType="Agents")
        self.assertFalse(res["OK"])

        agents = {
            "Agents": {
                "DataManagement": {
                    "FTS3Agent": {
                        "MEM": "0.3",
                        "Setup": True,
                        "PID": "18128",
                        "RunitStatus": "Run",
                        "Module": "CleanFTSDBAgent",
                        "Installed": True,
                        "VSZ": "375576",
                        "Timeup": "29841",
                        "CPU": "0.0",
                        "RSS": "55452",
                    }
                },
                "Framework": {
                    "ErrorMessageMonitor": {
                        "MEM": "0.3",
                        "Setup": True,
                        "PID": "2303",
                        "RunitStatus": "Run",
                        "Module": "ErrorMessageMonitor",
                        "Installed": True,
                        "VSZ": "380392",
                        "Timeup": "3380292",
                        "CPU": "0.0",
                        "RSS": "56172",
                    }
                },
                "System": {
                    "Off": {
                        "MEM": "0.3",
                        "Setup": True,
                        "PID": "---",
                        "RunitStatus": "Down",
                        "Module": "ErrorMessageMonitor",
                        "Installed": True,
                        "VSZ": "380392",
                        "Timeup": "3380292",
                        "CPU": "0.0",
                        "RSS": "56172",
                    }
                },
            }
        }
        agents["Agents"]["DataManagement"]["FTSAgent"] = {
            "Setup": False,
            "PID": 0,
            "RunitStatus": "Unknown",
            "Module": "FTSAgent",
            "Installed": False,
            "Timeup": 0,
        }

        self.restartAgent.sysAdminClient.getOverallStatus.return_value = S_OK(agents)
        res = self.restartAgent.getRunningInstances(instanceType="Agents")

        # only insalled agents with RunitStatus RUN should be returned
        self.assertTrue("FTSAgent" not in res["Value"])
        self.assertTrue("FTS3Agent" in res["Value"])
        self.assertTrue("ErrorMessageMonitor" in res["Value"])
        for agent in res["Value"]:
            self.assertTrue("PollingTime" in res["Value"][agent])
            self.assertTrue("LogFileLocation" in res["Value"][agent])
            self.assertTrue("PID" in res["Value"][agent])

    @patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.gConfig", new=MagicMock())
    def test_execute(self):
        """Test for the execute function."""
        self.restartAgent.sendNotification = MagicMock()

        agentOne = "FTS3Agent"
        agentTwo = "FTSAgent"
        agentOnePollingTime = 100
        agentTwoPollingTime = 200
        agentOneLogLoc = "/fake/loc1"
        agentTwoLogLoc = "/fake/loc2"
        agentOnePID = "12345"
        agentTwoPID = "54321"

        self.restartAgent.agents = {
            agentOne: {"PollingTime": agentOnePollingTime, "LogFileLocation": agentOneLogLoc, "PID": agentOnePID},
            agentTwo: {"PollingTime": agentTwoPollingTime, "LogFileLocation": agentTwoLogLoc, "PID": agentTwoPID},
        }

        self.restartAgent.checkAgent = MagicMock(side_effect=[S_OK(), S_ERROR()])
        self.restartAgent.componentControl = MagicMock(return_value=S_OK())

        res = self.restartAgent.execute()

        self.assertFalse(res["OK"])

        calls = [call(agentOne, self.restartAgent.agents[agentOne]), call(agentTwo, self.restartAgent.agents[agentTwo])]
        self.restartAgent.checkAgent.assert_has_calls(calls, any_order=True)

        # email notification should be sent at the end of every agent cycle
        self.restartAgent.sendNotification.assert_called()

    @patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.gConfig", new=MagicMock())
    def test_execute_2a(self):
        """Test for the execute function."""
        self.restartAgent.sendNotification = MagicMock()
        self.restartAgent.componentControl = MagicMock(return_value=S_ERROR("Stopped does not exist"))
        self.restartAgent.checkURLs = MagicMock(return_value=S_ERROR("SomeFailure"))
        res = self.restartAgent.execute()
        self.assertFalse(res["OK"])
        # email notification should be sent at the end of every agent cycle
        self.restartAgent.sendNotification.assert_called()

    @patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.gConfig", new=MagicMock())
    def test_execute_2b(self):
        """Test for the execute function."""
        self.restartAgent.sendNotification = MagicMock()
        self.restartAgent.componentControl = MagicMock(return_value=S_ERROR())
        self.restartAgent.checkURLs = MagicMock()
        res = self.restartAgent.execute()
        self.assertFalse(res["OK"])
        self.assertIn("Failure to control components ", self.restartAgent.errors)
        # email notification should be sent at the end of every agent cycle
        self.restartAgent.sendNotification.assert_called()
        self.restartAgent.checkURLs.assert_not_called()

    @patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.gConfig", new=MagicMock())
    def test_execute_2c(self):
        """Test for the execute function."""
        self.restartAgent.sendNotification = MagicMock()
        self.restartAgent.componentControl = MagicMock(return_value=S_OK())
        self.restartAgent.checkURLs = MagicMock(return_value=S_OK())
        res = self.restartAgent.execute()
        self.assertEqual([], self.restartAgent.errors)
        self.assertTrue(res["OK"])
        # email notification should be sent at the end of every agent cycle
        self.restartAgent.sendNotification.assert_called()
        self.restartAgent.checkURLs.assert_called_once()

    def test_check_agent(self):
        """Test for the checkAgent function."""
        self.restartAgent.getLastAccessTime = MagicMock()
        self.restartAgent.restartInstance = MagicMock(return_value=S_OK())

        agentName = "agentX"
        options = dict(PollingTime=MAA.HOUR, LogFileLocation="/fake/log/file", PID="12345")

        self.restartAgent.getLastAccessTime.return_value = S_ERROR()
        res = self.restartAgent.checkAgent(agentName, options)
        self.assertFalse(res["OK"])

        # agents with log file age less than max(pollingTime+Hour, 2 Hour) should not be restarted
        logAge = timedelta(hours=1)
        self.restartAgent.getLastAccessTime.return_value = S_OK(logAge)
        res = self.restartAgent.checkAgent(agentName, options)
        self.assertTrue(res["OK"])
        self.restartAgent.restartInstance.assert_not_called()

        # agents with log file age of more than max(pollingTime+Hour, 2 Hour) should be restarted
        logAge = timedelta(hours=3)
        self.restartAgent.restartAgents = False
        self.restartAgent.getLastAccessTime.return_value = S_OK(logAge)
        res = self.restartAgent.checkAgent(agentName, options)
        self.assertTrue(res["OK"])
        self.restartAgent.restartInstance.assert_called_once_with(12345, agentName, False)

    def test_check_executors(self):
        """Test the checkExecutor function."""
        self.restartAgent.getLastAccessTime = MagicMock()
        self.restartAgent.restartInstance = MagicMock(return_value=S_OK())
        self.restartAgent.checkForCheckingJobs = MagicMock()
        self.restartAgent.restartExecutors = True

        agentName = "executorX"
        currentLogLocation = "/fake/log/file"
        pid = "12345"
        options = dict(PID=pid, LogFileLocation=currentLogLocation)
        self.restartAgent.getLastAccessTime.return_value = S_ERROR()
        res = self.restartAgent.checkExecutor(agentName, options)
        self.assertFalse(res["OK"])

        # log file ok
        logAge = timedelta(hours=1)
        self.restartAgent.getLastAccessTime.return_value = S_OK(logAge)
        res = self.restartAgent.checkExecutor(agentName, options)
        self.assertTrue(res["OK"])
        self.restartAgent.restartInstance.assert_not_called()

        # log file too old, no checking jobs
        logAge = timedelta(hours=3)
        self.restartAgent.getLastAccessTime.return_value = S_OK(logAge)
        self.restartAgent.checkForCheckingJobs.return_value = S_OK("NO_CHECKING_JOBS")
        res = self.restartAgent.checkExecutor(agentName, options)
        self.assertTrue(res["OK"])
        self.restartAgent.restartInstance.assert_not_called()

        # log file too old, checking jobs failed
        logAge = timedelta(hours=3)
        self.restartAgent.getLastAccessTime.return_value = S_OK(logAge)
        self.restartAgent.checkForCheckingJobs.return_value = S_ERROR()
        res = self.restartAgent.checkExecutor(agentName, options)
        self.assertFalse(res["OK"])
        self.restartAgent.restartInstance.assert_not_called()

        # log file too old, checking jobs OK
        logAge = timedelta(hours=3)
        self.restartAgent.getLastAccessTime.return_value = S_OK(logAge)
        self.restartAgent.checkForCheckingJobs.return_value = S_OK("CHECKING_JOBS")
        res = self.restartAgent.checkExecutor(agentName, options)
        self.assertTrue(res["OK"])
        self.restartAgent.restartInstance.assert_called_once_with(int(pid), agentName, True)

        # log file too old, checking jobs OK, restart failed
        self.restartAgent.restartInstance = MagicMock(return_value=S_ERROR())
        logAge = timedelta(hours=3)
        self.restartAgent.getLastAccessTime.return_value = S_OK(logAge)
        self.restartAgent.checkForCheckingJobs.return_value = S_OK("CHECKING_JOBS")
        res = self.restartAgent.checkExecutor(agentName, options)
        self.assertFalse(res["OK"])
        self.restartAgent.restartInstance.assert_called_once_with(int(pid), agentName, True)

        # log file too old, checking jobs OK, restart disabled
        self.restartAgent.restartExecutors = False
        self.restartAgent.restartInstance = MagicMock(return_value=S_OK("NO_RESTART"))
        logAge = timedelta(hours=3)
        self.restartAgent.getLastAccessTime.return_value = S_OK(logAge)
        self.restartAgent.checkForCheckingJobs.return_value = S_OK("CHECKING_JOBS")
        res = self.restartAgent.checkExecutor(agentName, options)
        self.assertTrue(res["OK"])
        self.restartAgent.restartInstance.assert_called_once_with(int(pid), agentName, False)

    def test_check_services(self):
        """Test for checkServices function."""
        self.restartAgent.getLastAccessTime = MagicMock()
        self.restartAgent.restartInstance = MagicMock(return_value=S_OK())

        agentName = "serviceX"
        pid = "12345"
        options = dict(PID=pid, System="Skynet", Port=999)
        self.restartAgent.restartServices = False

        # service responds ok
        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.Client", new=clientMock(S_OK())):
            res = self.restartAgent.checkService(agentName, options)
        self.assertTrue(res["OK"])
        self.restartAgent.restartInstance.assert_not_called()

        # service responds not ok
        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.Client", new=clientMock(S_ERROR())):
            res = self.restartAgent.checkService(agentName, options)
        self.assertTrue(res["OK"])
        self.restartAgent.restartInstance.assert_called_once_with(int(pid), agentName, False)

        # service responds not ok
        self.restartAgent.restartInstance.reset_mock()
        self.restartAgent.restartInstance.return_value = S_OK("NO_RESTART")
        self.restartAgent.restartServices = False
        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.Client", new=clientMock(S_ERROR())):
            res = self.restartAgent.checkService(agentName, options)
        self.assertTrue(res["OK"])
        self.restartAgent.restartInstance.assert_called_once_with(int(pid), agentName, False)

        # service responds not ok, restart Failed
        self.restartAgent.restartInstance.reset_mock()
        self.restartAgent.restartServices = True
        self.restartAgent.restartInstance.return_value = S_ERROR()
        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.Client", new=clientMock(S_ERROR())):
            res = self.restartAgent.checkService(agentName, options)
        self.assertFalse(res["OK"])
        self.restartAgent.restartInstance.assert_called_once_with(int(pid), agentName, True)

    def test_get_last_access_time(self):
        """Test for the getLastAccessTime function."""
        self.agent.os.path.getmtime = MagicMock()
        self.agent.datetime = MagicMock()
        self.agent.datetime.now = MagicMock()
        self.agent.datetime.fromtimestamp = MagicMock()

        now = datetime.now()
        self.agent.datetime.now.return_value = now
        self.agent.datetime.fromtimestamp.return_value = now - timedelta(hours=1)

        res = self.restartAgent.getLastAccessTime("/fake/file")
        self.assertTrue(res["OK"])
        self.assertIsInstance(res["Value"], timedelta)
        self.assertEqual(res["Value"].seconds, 3600)

    def test_restartInstance(self):
        """Test restartInstance function."""
        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.psutil", new=self.getPSMock()):
            res = self.restartAgent.restartInstance(12345, "agentX", True)
        self.assertTrue(res["OK"])

        psMock = self.getPSMock()
        psMock.Process = MagicMock("RaisingProc")
        psMock.Error = psutil.Error
        psMock.Process.side_effect = psutil.Error()
        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.psutil", new=psMock):
            res = self.restartAgent.restartInstance(12345, "agentX", True)
        self.assertFalse(res["OK"])

        # restarting forbidden
        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.psutil", new=self.getPSMock()):
            res = self.restartAgent.restartInstance(12345, "agentFoo", True)
        self.assertTrue(res["OK"])
        assert res["Value"] == NO_RESTART

        # restarting disabled
        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.psutil", new=self.getPSMock()):
            res = self.restartAgent.restartInstance(12345, "agentY", False)
        self.assertTrue(res["OK"])
        assert res["Value"] == NO_RESTART

    def test_checkingJobs(self):
        """Test checkForCheckingJobs function."""
        self.restartAgent.jobMonClient = MagicMock()

        # failed to get jobs
        self.restartAgent.jobMonClient.getJobs.return_value = S_ERROR()
        self.assertFalse(self.restartAgent.checkForCheckingJobs("executor")["OK"])

        # No checking jobs
        self.restartAgent.jobMonClient.getJobs.return_value = S_OK(0)
        self.assertTrue(self.restartAgent.checkForCheckingJobs("executor")["OK"])
        self.assertEqual(self.restartAgent.checkForCheckingJobs("executor")["Value"], "NO_CHECKING_JOBS")

        # checking jobs
        self.restartAgent.jobMonClient.getJobs.return_value = S_OK({1: 1, 2: 2})
        self.assertTrue(self.restartAgent.checkForCheckingJobs("executor")["OK"])
        self.assertEqual(self.restartAgent.checkForCheckingJobs("executor")["Value"], "CHECKING_JOBS")

    def test_componentControl_1(self):
        """Fail to get status."""
        self.restartAgent.controlComponents = True
        self.restartAgent.errors = []
        self.restartAgent.sysAdminClient.getOverallStatus.return_value = S_ERROR()
        self.assertFalse(self.restartAgent.componentControl()["OK"])

    def test_componentControl_2(self):
        """Total Success."""
        self.restartAgent.controlComponents = True
        self.restartAgent.errors = []

        def god(*args, **_kwargs):
            """Mock getOptionsDict."""
            if "running" in args[0].lower():
                return S_OK({"Framework__StartMe": "", "Framework__Running": ""})
            if "stopped" in args[0].lower():
                return S_OK({"Framework__StopMe": "", "Framework__Stopped": "", "Framework__Unknown": ""})
            return S_ERROR()

        gConfigMock = MagicMock()
        gConfigMock.getOptionsDict.side_effect = god

        agents = {
            "Agents": {
                "Framework": {
                    "StopMe": {"RunitStatus": "Run"},
                    "StartMe": {"RunitStatus": "Down"},
                    "Running": {"RunitStatus": "Run"},
                    "Stopped": {"RunitStatus": "Down"},
                    "Uninstalled": {"RunitStatus": "Unknown"},
                }
            }
        }
        self.restartAgent.sysAdminClient.getOverallStatus.return_value = S_OK(agents)
        self.restartAgent.sysAdminClient.startComponent = MagicMock()
        self.restartAgent.sysAdminClient.stopComponent = MagicMock()

        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.gConfig", new=gConfigMock):
            res = self.restartAgent.componentControl()
        self.assertTrue(res["OK"])
        self.restartAgent.sysAdminClient.startComponent.assert_called_with("Framework", "StartMe")
        self.restartAgent.sysAdminClient.stopComponent.assert_called_with("Framework", "StopMe")
        self.assertIn("Unknown instance", self.restartAgent.errors[0])

    def test_componentControl_2b(self):
        """Total Success."""
        self.restartAgent.controlComponents = False
        self.restartAgent.errors = []

        def god(*args, **_kwargs):
            """Mock getOptionsDict."""
            if "running" in args[0].lower():
                return S_OK({"Framework__StartMe": "", "Framework__Running": ""})
            if "stopped" in args[0].lower():
                return S_OK({"Framework__StopMe": "", "Framework__Stopped": "", "Framework__Unknown": ""})
            return S_ERROR()

        gConfigMock = MagicMock()
        gConfigMock.getOptionsDict.side_effect = god

        agents = {
            "Agents": {
                "Framework": {
                    "StopMe": {"RunitStatus": "Run"},
                    "StartMe": {"RunitStatus": "Down"},
                    "Running": {"RunitStatus": "Run"},
                    "Stopped": {"RunitStatus": "Down"},
                    "Uninstalled": {"RunitStatus": "Unknown"},
                }
            }
        }
        self.restartAgent.sysAdminClient.getOverallStatus.return_value = S_OK(agents)
        self.restartAgent.sysAdminClient.startComponent = MagicMock()
        self.restartAgent.sysAdminClient.stopComponent = MagicMock()

        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.gConfig", new=gConfigMock):
            res = self.restartAgent.componentControl()
        self.assertTrue(res["OK"])
        self.restartAgent.sysAdminClient.startComponent.assert_not_called()
        self.restartAgent.sysAdminClient.stopComponent.assert_not_called()
        self.assertIn("Unknown instance", self.restartAgent.errors[0])
        self.assertIn("should be started", self.restartAgent.accounting["Framework__StartMe"]["Treatment"])
        self.assertIn("should be stopped", self.restartAgent.accounting["Framework__StopMe"]["Treatment"])

    def test_componentControl_3(self):
        """Failed to get options."""
        self.restartAgent.controlComponents = True
        self.restartAgent.errors = []

        def god(*args, **_kwargs):
            """Mock getOptionsDict."""
            if "running" in args[0].lower():
                return S_OK({"Framework__StopMe": "", "Framework__Stopped": "", "Framework__Unknown": ""})
            if "stopped" in args[0].lower():
                return S_ERROR("No Stopped")
            return S_ERROR()

        gConfigMock = MagicMock()
        gConfigMock.getOptionsDict.side_effect = god

        self.restartAgent.sysAdminClient.getOverallStatus.return_value = S_OK({})
        self.restartAgent.sysAdminClient.startComponent = MagicMock()
        self.restartAgent.sysAdminClient.stopComponent = MagicMock()

        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.gConfig", new=gConfigMock):
            res = self.restartAgent.componentControl()
        self.assertFalse(res["OK"])
        self.assertIn("No Stopped", res["Message"])

    def test_componentControl_4(self):
        """Bad host config."""
        self.restartAgent.controlComponents = True
        self.restartAgent.errors = []

        def god(*args, **_kwargs):
            """Mock getOptionsDict."""
            if "running" in args[0].lower():
                return S_OK({"Framework__Twice": ""})
            if "stopped" in args[0].lower():
                return S_OK({"Framework__Twice": ""})
            return S_ERROR()

        gConfigMock = MagicMock()
        gConfigMock.getOptionsDict.side_effect = god

        self.restartAgent.sysAdminClient.getOverallStatus.return_value = S_OK({})
        self.restartAgent.sysAdminClient.startComponent = MagicMock()
        self.restartAgent.sysAdminClient.stopComponent = MagicMock()

        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.gConfig", new=gConfigMock):
            res = self.restartAgent.componentControl()
        self.assertFalse(res["OK"])
        self.assertIn("Bad host configuration", res["Message"])

    def test_checkURLs_1(self):
        """Success."""
        theTornadoPort = "000043000"
        self.restartAgent.errors = []
        self.restartAgent.accounting.clear()
        host = "foo.server"
        prot = ["", "dips", "dips", "https"]
        port = ["", "1001", "1002", theTornadoPort]
        urls, tempurls, newurls = [], [], []
        for i in [1, 2]:
            urls.append("%(prot)s://%(host)s:%(port)s/Sys/Serv%(i)s" % dict(i=i, host=host, port=port[i], prot=prot[i]))
        for i in [1]:
            tempurls.append(
                "%(prot)s://%(host)s:%(port)s/Sys/Serv%(i)s" % dict(i=i, host=host, port=port[i], prot=prot[i])
            )
        for i in [1, 3]:
            newurls.append(
                "%(prot)s://%(host)s:%(port)s/Sys/Serv%(i)s" % dict(i=i, host=host, port=port[i], prot=prot[i])
            )

        def gVal(*args, **_kwargs):
            """Mock getValue."""
            if "Tornado" in args[0]:
                return theTornadoPort
            if "PollingTime" in args[0]:
                return 365
            if "Port" in args[0]:
                return args[1] if "Serv3" in args[0] else "100" + args[0].rsplit("/Serv", 1)[1].split("/")[0]
            if "URLs" in args[0]:
                return urls
            if "Protocol" in args[0]:
                return "https" if "Serv3" in args[0] else args[1]
            else:
                assert False, "Unknown config option requested %s" % args[0]

        gConfigMock = MagicMock()
        gConfigMock.getValue.side_effect = gVal
        services = {
            "Services": {
                "Sys": {
                    "Serv1": {
                        "Setup": True,
                        "PID": "18128",
                        "Port": "1001",
                        "RunitStatus": "Run",
                        "Module": "Serv",
                        "Installed": True,
                    },
                    "Serv2": {
                        "Setup": True,
                        "PID": "18128",
                        "Port": "1002",
                        "RunitStatus": "Down",
                        "Module": "Serv",
                        "Installed": True,
                    },
                    "Serv3": {
                        "Setup": True,
                        "PID": "18128",
                        "RunitStatus": "Run",
                        "Protocol": "https",
                        "Module": "Serv",
                        "Installed": True,
                    },
                    "SystemAdministrator": {
                        "Setup": True,
                        "PID": "18128",
                        "Port": "1003",
                        "RunitStatus": "Run",
                        "Module": "Serv",
                        "Installed": True,
                    },
                }
            }
        }

        self.restartAgent.sysAdminClient.getOverallStatus.return_value = S_OK(services)

        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.gConfig", new=gConfigMock), patch(
            "DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.socket.gethostname", return_value=host
        ), patch(
            "DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.PathFinder.getSystemInstance",
            return_value="Decertification",
        ), patch(
            "DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.PathFinder.getComponentSection",
            side_effect=mockComponentSection,
        ), patch(
            "DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.PathFinder.getSystemURLSection",
            side_effect=mockURLSection,
        ):
            res = self.restartAgent.checkURLs()
        self.assertTrue(res["OK"])
        self.restartAgent.csAPI.modifyValue.assert_has_calls(
            [
                call("/Systems/Sys/Production/URLs/Serv", ",".join(tempurls)),
                call("/Systems/Sys/Production/URLs/Serv", ",".join(newurls)),
            ],
            any_order=False,
        )
        assert self.restartAgent._tornadoPort == theTornadoPort

    def test_checkURLs_2(self):
        """Test commit to CS."""
        self.restartAgent.errors = []
        self.restartAgent.accounting.clear()
        self.restartAgent.csAPI.csModified = True
        self.restartAgent.commitURLs = True
        self.restartAgent.sysAdminClient.getOverallStatus.return_value = S_OK(dict(Services={}))

        self.restartAgent.csAPI.commit = MagicMock(return_value=S_ERROR("Nope"))
        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.gConfig", new=MagicMock()), patch(
            "DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.PathFinder.getSystemInstance",
            return_value="Decertification",
        ), patch(
            "DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.PathFinder.getComponentSection",
            side_effect=mockComponentSection,
        ):
            res = self.restartAgent.checkURLs()
        self.assertFalse(res["OK"])
        self.assertIn("Failed to commit", res["Message"])
        self.assertIn("Commit to CS failed", self.restartAgent.errors[0])

        self.restartAgent.csAPI.commit = MagicMock(return_value=S_OK())
        with patch("DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.gConfig", new=MagicMock()), patch(
            "DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.PathFinder.getSystemInstance",
            return_value="Decertification",
        ), patch(
            "DIRAC.FrameworkSystem.Agent.ComponentSupervisionAgent.PathFinder.getComponentSection",
            side_effect=mockComponentSection,
        ):

            res = self.restartAgent.checkURLs()
        self.assertTrue(res["OK"])


if __name__ == "__main__":
    SUITE = unittest.defaultTestLoader.loadTestsFromTestCase(TestComponentSupervisionAgent)
    TESTRESULT = unittest.TextTestRunner(verbosity=3).run(SUITE)



########################################################################
# File: MetaQuery.py
# Author: A.T.
# Date: 24.02.2015
# $HeadID$
########################################################################

""" Utilities for managing metadata based queries
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id$"

from DIRAC import S_OK, S_ERROR
import DIRAC.Core.Utilities.Time as Time

import json

FILE_STANDARD_METAKEYS = {
    "SE": "VARCHAR",
    "CreationDate": "DATETIME",
    "ModificationDate": "DATETIME",
    "LastAccessDate": "DATETIME",
    "User": "VARCHAR",
    "Group": "VARCHAR",
    "Path": "VARCHAR",
    "Name": "VARCHAR",
    "FileName": "VARCHAR",
    "CheckSum": "VARCHAR",
    "GUID": "VARCHAR",
    "UID": "INTEGER",
    "GID": "INTEGER",
    "Size": "INTEGER",
    "Status": "VARCHAR",
}

FILES_TABLE_METAKEYS = {
    "Name": "FileName",
    "FileName": "FileName",
    "Size": "Size",
    "User": "UID",
    "Group": "GID",
    "UID": "UID",
    "GID": "GID",
    "Status": "Status",
}

FILEINFO_TABLE_METAKEYS = {
    "GUID": "GUID",
    "CheckSum": "CheckSum",
    "CreationDate": "CreationDate",
    "ModificationDate": "ModificationDate",
    "LastAccessDate": "LastAccessDate",
}


class MetaQuery(object):
    def __init__(self, queryDict=None, typeDict=None):

        self.__metaQueryDict = {}
        if queryDict is not None:
            self.__metaQueryDict = queryDict
        self.__metaTypeDict = {}
        if typeDict is not None:
            self.__metaTypeDict = typeDict

    def setMetaQuery(self, queryList, metaTypeDict=None):
        """Create the metadata query out of the command line arguments"""
        if metaTypeDict is not None:
            self.__metaTypeDict = metaTypeDict
        metaDict = {}
        contMode = False
        value = ""
        for arg in queryList:
            if not contMode:
                operation = ""
                for op in [">=", "<=", ">", "<", "!=", "="]:
                    if op in arg:
                        operation = op
                        break
                if not operation:
                    return S_ERROR("Illegal query element %s" % arg)

                name, value = arg.split(operation)
                if name not in self.__metaTypeDict:
                    return S_ERROR("Metadata field %s not defined" % name)

                mtype = self.__metaTypeDict[name]
            else:
                value += " " + arg
                value = value.replace(contMode, "")
                contMode = False

            if value[0] in ['"', "'"] and value[-1] not in ['"', "'"]:
                contMode = value[0]
                continue

            if "," in value:
                valueList = [x.replace("'", "").replace('"', "") for x in value.split(",")]
                mvalue = valueList
                if mtype[0:3].lower() == "int":
                    mvalue = [int(x) for x in valueList if x not in ["Missing", "Any"]]
                    mvalue += [x for x in valueList if x in ["Missing", "Any"]]
                if mtype[0:5].lower() == "float":
                    mvalue = [float(x) for x in valueList if x not in ["Missing", "Any"]]
                    mvalue += [x for x in valueList if x in ["Missing", "Any"]]
                if operation == "=":
                    operation = "in"
                if operation == "!=":
                    operation = "nin"
                mvalue = {operation: mvalue}
            else:
                mvalue = value.replace("'", "").replace('"', "")
                if value not in ["Missing", "Any"]:
                    if mtype[0:3].lower() == "int":
                        mvalue = int(value)
                    if mtype[0:5].lower() == "float":
                        mvalue = float(value)
                if operation != "=":
                    mvalue = {operation: mvalue}

            if name in metaDict:
                if isinstance(metaDict[name], dict):
                    if isinstance(mvalue, dict):
                        op, value = list(mvalue.items())[0]
                        if op in metaDict[name]:
                            if isinstance(metaDict[name][op], list):
                                if isinstance(value, list):
                                    metaDict[name][op] = list(set(metaDict[name][op] + value))
                                else:
                                    metaDict[name][op] = list(set(metaDict[name][op].append(value)))
                            else:
                                if isinstance(value, list):
                                    metaDict[name][op] = list(set([metaDict[name][op]] + value))
                                else:
                                    metaDict[name][op] = list(set([metaDict[name][op], value]))
                        else:
                            metaDict[name].update(mvalue)
                    else:
                        if isinstance(mvalue, list):
                            metaDict[name].update({"in": mvalue})
                        else:
                            metaDict[name].update({"=": mvalue})
                elif isinstance(metaDict[name], list):
                    if isinstance(mvalue, dict):
                        metaDict[name] = {"in": metaDict[name]}
                        metaDict[name].update(mvalue)
                    elif isinstance(mvalue, list):
                        metaDict[name] = list(set((metaDict[name] + mvalue)))
                    else:
                        metaDict[name] = list(set(metaDict[name].append(mvalue)))
                else:
                    if isinstance(mvalue, dict):
                        metaDict[name] = {"=": metaDict[name]}
                        metaDict[name].update(mvalue)
                    elif isinstance(mvalue, list):
                        metaDict[name] = list(set([metaDict[name]] + mvalue))
                    else:
                        metaDict[name] = list(set([metaDict[name], mvalue]))
            else:
                metaDict[name] = mvalue

        self.__metaQueryDict = metaDict
        return S_OK(metaDict)

    def getMetaQuery(self):

        return self.__metaQueryDict

    def getMetaQueryAsJson(self):

        return json.dumps(self.__metaQueryDict)

    def applyQuery(self, userMetaDict):
        """Return a list of tuples with tables and conditions to locate files for a given user Metadata"""

        def getOperands(value):
            if isinstance(value, list):
                return [("in", value)]
            elif isinstance(value, dict):
                resultList = []
                for operation, operand in value.items():
                    resultList.append((operation, operand))
                return resultList
            else:
                return [("=", value)]

        def getTypedValue(value, mtype):
            if mtype[0:3].lower() == "int":
                return int(value)
            elif mtype[0:5].lower() == "float":
                return float(value)
            elif mtype[0:4].lower() == "date":
                return Time.fromString(value)
            else:
                return value

        for meta, value in self.__metaQueryDict.items():

            # Check if user dict contains all the requested meta data
            userValue = userMetaDict.get(meta, None)
            if userValue is None:
                if str(value).lower() == "missing":
                    continue
                else:
                    return S_OK(False)
            elif str(value).lower() == "any":
                continue

            mtype = self.__metaTypeDict[meta]
            try:
                userValue = getTypedValue(userValue, mtype)
            except ValueError:
                return S_ERROR("Illegal type for metadata %s: %s in user data" % (meta, str(userValue)))

            # Check operations
            for operation, operand in getOperands(value):
                try:
                    if isinstance(operand, list):
                        typedValue = [getTypedValue(x, mtype) for x in operand]
                    else:
                        typedValue = getTypedValue(operand, mtype)
                except ValueError:
                    return S_ERROR("Illegal type for metadata %s: %s in filter" % (meta, str(operand)))

                # Apply query operation
                if operation in [">", "<", ">=", "<="]:
                    if isinstance(typedValue, list):
                        return S_ERROR("Illegal query: list of values for comparison operation")
                    elif operation == ">" and typedValue >= userValue:
                        return S_OK(False)
                    elif operation == "<" and typedValue <= userValue:
                        return S_OK(False)
                    elif operation == ">=" and typedValue > userValue:
                        return S_OK(False)
                    elif operation == "<=" and typedValue < userValue:
                        return S_OK(False)
                elif operation == "in" or operation == "=":
                    if isinstance(typedValue, list) and userValue not in typedValue:
                        return S_OK(False)
                    elif not isinstance(typedValue, list) and userValue != typedValue:
                        return S_OK(False)
                elif operation == "nin" or operation == "!=":
                    if isinstance(typedValue, list) and userValue in typedValue:
                        return S_OK(False)
                    elif not isinstance(typedValue, list) and userValue == typedValue:
                        return S_OK(False)

        return S_OK(True)



""" General Message Queue Interface to create Consumers and Producers
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from DIRAC import gLogger, S_OK
from DIRAC.Resources.MessageQueue.MQProducer import MQProducer
from DIRAC.Resources.MessageQueue.MQConsumer import MQConsumer
from DIRAC.Resources.MessageQueue.MQConnectionManager import MQConnectionManager
from DIRAC.Resources.MessageQueue.Utilities import getMQParamsFromCS
from DIRAC.Resources.MessageQueue.Utilities import generateDefaultCallback

__RCSID__ = "$Id$"

connectionManager = MQConnectionManager()  # To manage the active MQ connections.


def createConsumer(mqURI, callback=generateDefaultCallback()):
    """
    Function creates MQConsumer. All parameters are taken from the
    Configuration Service based on the mqURI value.

    Args:
      mqURI(str):Pseudo URI identifing MQ service. It has the following format
                mqConnection::DestinationType::DestinationName
                e.g. blabla.cern.ch::Queues::MyQueue1
      callback: callback function that can be used to process the incoming messages

    Returns:
      S_OK/S_ERROR: with the consumer object in S_OK.

    """
    result = _setupConnection(mqURI=mqURI, mType="consumer")
    if not result["OK"]:
        return result
    return S_OK(MQConsumer(mqManager=connectionManager, mqURI=mqURI, consumerId=result["Value"], callback=callback))


def createProducer(mqURI):
    """
    Function creates MQProducer. All parameters are taken from
    the Configuration Service based on the mqURI value.

    Args:
      mqURI(str):Pseudo URI identifing MQ service. It has the following format
                mqConnection::DestinationType::DestinationName
                e.g. blabla.cern.ch::Queues::MyQueue1
    Returns:
      S_OK/S_ERROR: with the producer object in S_OK.
    """
    result = _setupConnection(mqURI=mqURI, mType="producer")
    if not result["OK"]:
        return result
    return S_OK(MQProducer(mqManager=connectionManager, mqURI=mqURI, producerId=result["Value"]))


def _setupConnection(mqURI, mType):
    """Function sets up the active MQ connection. All parameters are taken
        from the Configuration Service based on the mqURI
        value and the messenger Type mType.

    Args:
      mqURI(str):Pseudo URI identifing the MQ service. It has the following format:
                mqConnection::DestinationType::DestinationName
                e.g. blabla.cern.ch::Queues::MyQueue1
      mType(str): 'consumer' or 'producer'
    Returns:
      S_OK/S_ERROR: with the value of the messenger Id ( e.g. 'consumer4' ) in S_OK.
    """
    result = getMQParamsFromCS(mqURI=mqURI)
    if not result["OK"]:
        return result
    params = result["Value"]
    return connectionManager.startConnection(mqURI, params, mType)



""" ModuleBase - contains the base class for workflow modules. Defines several common utility methods.

    The modules defined within this package are developed in a way to be executed by a DIRAC.Core.Worfklow.Worfklow.
    In particular, a DIRAC.Core.Workflow.Worfklow object will only call the "execute" function, that is defined here.

    These modules, inspired by the LHCb experience, give the possibility to define simple user and production jobs.
    Many VOs might want to extend this package. And actually, for some cases, it will be necessary. For example,
    defining the LFN output at runtime (within the "UploadOutputs" module is a VO specific operation

    The DIRAC APIs are used to create Jobs that make use of these modules.
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import six
import os
import copy

from DIRAC import S_OK, S_ERROR, gLogger
from DIRAC.ConfigurationSystem.Client.Helpers.Operations import Operations
from DIRAC.WorkloadManagementSystem.Client.JobReport import JobReport
from DIRAC.TransformationSystem.Client.FileReport import FileReport
from DIRAC.RequestManagementSystem.Client.Request import Request
from DIRAC.RequestManagementSystem.private.RequestValidator import RequestValidator
from DIRAC.DataManagementSystem.Client.DataManager import DataManager


class ModuleBase(object):
    """Base class for Modules - works only within DIRAC workflows

    This module, inheriting by "object", can use cooperative methods, very useful here.
    """

    #############################################################################

    def __init__(self, loggerIn=None):
        """Initialization of module base.

        loggerIn is a logger object that can be passed so that the logging will be more clear.
        """

        if not loggerIn:
            self.log = gLogger.getSubLogger("ModuleBase")
        else:
            self.log = loggerIn

        # These 2 are used in many places, so it's good to have them available here.
        self.opsH = Operations()
        self.dm = DataManager()

        # Some job parameters
        self.production_id = 0
        self.prod_job_id = 0
        self.jobID = 0
        self.step_number = 0
        self.step_id = 0
        self.jobType = ""
        self.executable = ""
        self.command = None

        self.workflowStatus = {}
        self.stepStatus = {}
        self.workflow_commons = {}
        self.step_commons = {}

        self.applicationName = ""
        self.applicationVersion = ""
        self.outputDataFileMask = []
        self.stepName = ""
        self.stepInputData = []
        self.applicationLog = "applicationLog.txt"

        self.appSteps = []
        self.inputDataList = []
        self.InputData = []
        self.inputDataType = ""

        # These are useful objects (see the getFileReporter(), getJobReporter() and getRequestContainer() functions)
        self.fileReport = None
        self.jobReport = None
        self.request = None

    #############################################################################

    def execute(self):
        """Function called by all super classes. This is the only function that Workflow will call automatically.

        The design adopted here is that all the modules are inheriting from this class,
        and will NOT override this function. Instead, the inherited modules will override the following functions:
        _resolveInputVariables()
        _initialize()
        _setCommand()
        _executeCommand()
        _execute()
        that are called here exactly in this order.
        Each implementation of these functions, in the subclasses, should never return S_OK, S_ERROR.
        This choice has been made for convenience of coding, and for the high level of inheritance implemented here.
        Instead, they should return:
        - None when no issues arise
        - a RuntimeError exception when there are issues
        - a GracefulTermination exception (defined also here) when the module should be terminated gracefully

        The various parameters in input to this method are used almost only for testing purposes.
        """

        if not self.production_id:
            # self.PRODUCTION_ID is always set by the workflow
            self.production_id = int(self.workflow_commons["PRODUCTION_ID"])

        if not self.prod_job_id:
            # self.JOB_ID is set by the workflow, but this is not the WMS job id, but the transformation (production) task id
            self.prod_job_id = int(self.workflow_commons["JOB_ID"])

        if not self.jobID:
            # this is the real wms job ID
            self.jobID = int(os.environ.get("JOBID", self.jobID))

        if not self.step_number:
            # self.STEP_NUMBER is always set by the workflow
            self.step_number = int(self.STEP_NUMBER)  # pylint: disable=no-member

        if not self.step_id:
            self.step_id = "%d_%d_%d" % (self.production_id, self.prod_job_id, self.step_number)

        try:
            # This is what has to be extended in the modules
            self._resolveInputVariables()
            self._initialize()
            self._setCommand()
            self._executeCommand()
            self._execute()
            self._finalize()

        # If everything is OK
        except GracefulTermination as status:
            self.setApplicationStatus(status)
            self.log.info(status)
            return S_OK(status)

        # This catches everything that is voluntarily thrown within the modules, so an error
        except RuntimeError as rte:
            if len(rte.args) > 1:
                # In this case the RuntimeError is supposed to return in rte[1] an error code (possibly from DErrno)
                self.log.error(rte.args[0])
                self.setApplicationStatus(rte.args[0])
                return S_ERROR(rte.args[1], rte.args[0])  # rte[1] should be an error code

            # If we are here it is just a string
            self.log.error(rte)
            self.setApplicationStatus(rte)
            return S_ERROR(rte)

        # This catches everything that is not voluntarily thrown (here, really writing an exception)
        except BaseException as exc:  # pylint: disable=broad-except
            self.log.exception(exc)
            self.setApplicationStatus(exc)
            return S_ERROR(exc)

        finally:
            self.finalize()

    def _resolveInputVariables(self):
        """By convention the module input parameters are resolved here.
        fileReport, jobReport, and request objects are instantiated/recorded here.

        This will also call the resolution of the input workflow.
        The resolution of the input step should instead be done on a step basis.

        NB: Never forget to call this base method when extending it.
        """

        self.log.verbose("workflow_commons = ", self.workflow_commons)
        self.log.verbose("step_commons = ", self.step_commons)

        if not self.fileReport:
            self.fileReport = self._getFileReporter()
        if not self.jobReport:
            self.jobReport = self._getJobReporter()
        if not self.request:
            self.request = self._getRequestContainer()

        self._resolveInputWorkflow()

    def _initialize(self):
        """TBE

        For initializing the module, whatever operation this can be
        """
        pass

    def _setCommand(self):
        """TBE

        For "executors" modules, set the command to be used in the self.command variable.
        """
        pass

    def _executeCommand(self):
        """TBE

        For "executors" modules, executes self.command as set in the _setCommand() method
        """
        pass

    def _execute(self):
        """TBE

        Executes, whatever this means for the module implementing it
        """
        pass

    def _finalize(self, status=""):
        """TBE

        By default, the module finalizes correctly
        """

        if not status:
            status = "%s correctly finalized" % str(self.__class__)

        raise GracefulTermination(status)

    #############################################################################

    def finalize(self):
        """Just finalizing the module execution by flushing the logs. This will be done always."""

        self.log.info("===== Terminating " + str(self.__class__) + " ===== ")

    #############################################################################

    def _getJobReporter(self):
        """just return the job reporter (object, always defined by dirac-jobexec)"""

        if "JobReport" in self.workflow_commons:
            return self.workflow_commons["JobReport"]
        jobReport = JobReport(self.jobID)
        self.workflow_commons["JobReport"] = jobReport
        return jobReport

    #############################################################################

    def _getFileReporter(self):
        """just return the file reporter (object)"""

        if "FileReport" in self.workflow_commons:
            return self.workflow_commons["FileReport"]
        fileReport = FileReport()
        self.workflow_commons["FileReport"] = fileReport
        return fileReport

    #############################################################################

    def _getRequestContainer(self):
        """just return the RequestContainer reporter (object)"""

        if "Request" in self.workflow_commons:
            return self.workflow_commons["Request"]
        request = Request()
        self.workflow_commons["Request"] = request
        return request

    #############################################################################

    def _resolveInputWorkflow(self):
        """Resolve the input variables that are in the workflow_commons"""

        self.jobType = self.workflow_commons.get("JobType", self.jobType)

        self.InputData = ""
        if "InputData" in self.workflow_commons and self.workflow_commons["InputData"]:
            self.InputData = self.workflow_commons["InputData"]

        if "ParametricInputData" in self.workflow_commons:
            pID = copy.deepcopy(self.workflow_commons["ParametricInputData"])
            if pID:
                if isinstance(pID, list):
                    pID = ";".join(pID)
                #      self.InputData += ';' + pID
                self.InputData = pID
                self.InputData = self.InputData.rstrip(";")

        if self.InputData == ";":
            self.InputData = ""

        self.inputDataList = [lfn.strip("LFN:") for lfn in self.InputData.split(";") if lfn]

        self.appSteps = self.workflow_commons.get("appSteps", self.appSteps)

        if "outputDataFileMask" in self.workflow_commons:
            self.outputDataFileMask = self.workflow_commons["outputDataFileMask"]
            if isinstance(self.outputDataFileMask, six.string_types):
                self.outputDataFileMask = [
                    i.lower().strip() for i in self.outputDataFileMask.split(";")
                ]  # pylint: disable=no-member

    #############################################################################

    def _resolveInputStep(self):
        """Resolve the input variables for an application step"""

        self.stepName = self.step_commons["STEP_INSTANCE_NAME"]

        self.executable = self.step_commons.get("executable", "Unknown")

        self.applicationName = self.step_commons.get("applicationName", "Unknown")

        self.applicationVersion = self.step_commons.get("applicationVersion", "Unknown")

        self.applicationLog = self.step_commons.get(
            "applicationLog", self.step_commons.get("logFile", self.applicationLog)
        )

        self.inputDataType = self.step_commons.get("inputDataType", self.inputDataType)

        stepInputData = []
        if "inputData" in self.step_commons and self.step_commons["inputData"]:
            stepInputData = self.step_commons["inputData"]
        elif self.InputData:
            stepInputData = copy.deepcopy(self.InputData)
        if stepInputData:
            stepInputData = self._determineStepInputData(
                stepInputData,
            )
            self.stepInputData = [sid.strip("LFN:") for sid in stepInputData]

    #############################################################################

    def _determineStepInputData(self, inputData):
        """determine the input data for the step"""
        if inputData == "previousStep":
            stepIndex = self.appSteps.index(self.stepName)
            previousStep = self.appSteps[stepIndex - 1]

            stepInputData = []
            for outputF in self.workflow_commons["outputList"]:
                try:
                    if (
                        outputF["stepName"] == previousStep
                        and outputF["outputDataType"].lower() == self.inputDataType.lower()
                    ):
                        stepInputData.append(outputF["outputDataName"])
                except KeyError:
                    raise RuntimeError("Can't find output of step %s" % previousStep)

            return stepInputData

        return [x.strip("LFN:") for x in inputData.split(";")]

    #############################################################################

    def setApplicationStatus(self, status, sendFlag=True):
        """Wraps around setJobApplicationStatus of state update client"""
        if not self._WMSJob():
            return 0  # e.g. running locally prior to submission

        if self._checkWFAndStepStatus(noPrint=True):
            # The application status won't be updated in case the workflow or the step is failed already
            if not isinstance(status, six.string_types):
                status = str(status)
            self.log.verbose("setJobApplicationStatus(%s, %s)" % (self.jobID, status))
            jobStatus = self.jobReport.setApplicationStatus(status, sendFlag)
            if not jobStatus["OK"]:
                self.log.warn(jobStatus["Message"])

    #############################################################################

    def _WMSJob(self):
        """Check if this job is running via WMS"""
        return True if self.jobID else False

    #############################################################################

    def _enableModule(self):
        """Enable module if it's running via WMS"""
        if not self._WMSJob():
            self.log.info("No WMS JobID found, disabling module via control flag")
            return False
        else:
            self.log.verbose("Found WMS JobID = %d" % self.jobID)
            return True

    #############################################################################

    def _checkWFAndStepStatus(self, noPrint=False):
        """Check the WF and Step status"""
        if not self.workflowStatus["OK"] or not self.stepStatus["OK"]:
            if not noPrint:
                self.log.info("Skip this module, failure detected in a previous step :")
                self.log.info("Workflow status : %s" % (self.workflowStatus))
                self.log.info("Step Status : %s" % (self.stepStatus))
            return False
        else:
            return True

    #############################################################################

    def setJobParameter(self, name, value, sendFlag=True):
        """Wraps around setJobParameter of state update client"""
        if not self._WMSJob():
            return 0  # e.g. running locally prior to submission

        self.log.verbose("setJobParameter(%d,%s,%s)" % (self.jobID, name, value))

        jobParam = self.jobReport.setJobParameter(str(name), str(value), sendFlag)
        if not jobParam["OK"]:
            self.log.warn(jobParam["Message"])

    #############################################################################

    def getCandidateFiles(self, outputList, outputLFNs, fileMask, stepMask=""):
        """Returns list of candidate files to upload, check if some outputs are missing.

        outputList has the following structure:
          [ {'outputDataType':'','outputDataSE':'','outputDataName':''} , {...} ]

        outputLFNs is the list of output LFNs for the job

        fileMask is the output file extensions to restrict the outputs to

        returns dictionary containing type, SE and LFN for files restricted by mask
        """
        fileInfo = {}

        for outputFile in outputList:
            if "outputDataType" in outputFile and "outputDataSE" in outputFile and "outputDataName" in outputFile:
                fname = outputFile["outputDataName"]
                fileSE = outputFile["outputDataSE"]
                fileType = outputFile["outputDataType"]
                fileInfo[fname] = {"type": fileType, "workflowSE": fileSE}
            else:
                self.log.error("Ignoring malformed output data specification", str(outputFile))

        for lfn in outputLFNs:
            if os.path.basename(lfn) in fileInfo.keys():
                fileInfo[os.path.basename(lfn)]["lfn"] = lfn
                self.log.verbose("Found LFN %s for file %s" % (lfn, os.path.basename(lfn)))

        # check local existance
        self._checkLocalExistance(list(fileInfo))

        # Select which files have to be uploaded: in principle all
        candidateFiles = self._applyMask(fileInfo, fileMask, stepMask)

        # Sanity check all final candidate metadata keys are present (return S_ERROR if not)
        self._checkSanity(candidateFiles)

        return candidateFiles

    #############################################################################

    def _applyMask(self, candidateFilesIn, fileMask, stepMask):
        """Select which files have to be uploaded: in principle all"""
        candidateFiles = copy.deepcopy(candidateFilesIn)

        if fileMask and not isinstance(fileMask, list):
            fileMask = [fileMask]
        if isinstance(stepMask, int):
            stepMask = str(stepMask)
        if stepMask and not isinstance(stepMask, list):
            stepMask = [stepMask]

        if fileMask and fileMask != [""]:
            for fileName, metadata in list(candidateFiles.items()):
                if metadata["type"].lower() not in fileMask:  # and ( fileName.split( '.' )[-1] not in fileMask ) ):
                    del candidateFiles[fileName]
                    self.log.info(
                        "Output file %s was produced but will not be treated (fileMask is %s)"
                        % (fileName, ", ".join(fileMask))
                    )
        else:
            self.log.info("No outputDataFileMask provided, the files with all the extensions will be considered")

        if stepMask and stepMask != [""]:
            # This supposes that the LFN contains the step ID
            for fileName, metadata in list(candidateFiles.items()):
                if fileName.split("_")[-1].split(".")[0] not in stepMask:
                    del candidateFiles[fileName]
                    self.log.info(
                        "Output file %s was produced but will not be treated (stepMask is %s)"
                        % (fileName, ", ".join(stepMask))
                    )
        else:
            self.log.info("No outputDataStep provided, the files output of all the steps will be considered")

        return candidateFiles

    #############################################################################

    def _checkSanity(self, candidateFiles):
        """Sanity check all final candidate metadata keys are present"""

        notPresentKeys = []

        mandatoryKeys = ["type", "workflowSE", "lfn"]  # filedict is used for requests
        for fileName, metadata in candidateFiles.items():
            for key in mandatoryKeys:
                if key not in metadata:
                    notPresentKeys.append((fileName, key))

        if notPresentKeys:
            for fileName_keys in notPresentKeys:
                self.log.error("File %s has missing %s" % (fileName_keys[0], fileName_keys[1]))
            raise ValueError

    #############################################################################

    def _checkLocalExistance(self, fileList):
        """Check that the list of output files are present locally"""

        notPresentFiles = []

        for fileName in fileList:
            if not os.path.exists(fileName):
                notPresentFiles.append(fileName)

        if notPresentFiles:
            self.log.error("Output data file list %s does not exist locally" % notPresentFiles)
            raise os.error

    #############################################################################

    def generateFailoverFile(self):
        """Retrieve the accumulated reporting request, and produce a JSON file that is consumed by the JobWrapper"""
        reportRequest = None
        result = self.jobReport.generateForwardDISET()
        if not result["OK"]:
            self.log.warn("Could not generate Operation for job report with result:\n%s" % (result))
        else:
            reportRequest = result["Value"]
        if reportRequest:
            self.log.info("Populating request with job report information")
            self.request.addOperation(reportRequest)

        accountingReport = None
        if "AccountingReport" in self.workflow_commons:
            accountingReport = self.workflow_commons["AccountingReport"]
        if accountingReport:
            result = accountingReport.commit()
            if not result["OK"]:
                self.log.error("!!! Both accounting and RequestDB are down? !!!")
                return result

        if len(self.request):
            isValid = RequestValidator().validate(self.request)
            if not isValid["OK"]:
                raise RuntimeError("Failover request is not valid: %s" % isValid["Message"])
            else:
                requestJSON = self.request.toJSON()
                if requestJSON["OK"]:
                    self.log.info("Creating failover request for deferred operations for job %d" % self.jobID)
                    request_string = str(requestJSON["Value"])
                    self.log.debug(request_string)
                    # Write out the request string
                    fname = "%d_%d_request.json" % (self.production_id, self.prod_job_id)
                    jsonFile = open(fname, "w")
                    jsonFile.write(request_string)
                    jsonFile.close()
                    self.log.info("Created file containing failover request %s" % fname)
                    result = self.request.getDigest()
                    if result["OK"]:
                        self.log.info("Digest of the request: %s" % result["Value"])
                    else:
                        self.log.error("No digest? That's not sooo important, anyway:", result["Message"])
                else:
                    raise RuntimeError(requestJSON["Message"])

    #############################################################################


#############################################################################


class GracefulTermination(Exception):
    pass


#############################################################################



########################################################################
# File: RemoveReplica.py
# Author: Krzysztof.Ciba@NOSPAMgmail.com
# Date: 2013/03/25 07:45:06
########################################################################

""" :mod: RemoveReplica

    ===================

    .. module: RemoveReplica

    :synopsis: removeReplica operation handler

    .. moduleauthor:: Krzysztof.Ciba@NOSPAMgmail.com

    RemoveReplica operation handler
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

__RCSID__ = "$Id $"

# #
# @file RemoveReplica.py
# @author Krzysztof.Ciba@NOSPAMgmail.com
# @date 2013/03/25 07:45:17
# @brief Definition of RemoveReplica class.

# # imports
import os

# # from DIRAC
from DIRAC import S_OK
from DIRAC.FrameworkSystem.Client.MonitoringClient import gMonitor
from DIRAC.DataManagementSystem.Agent.RequestOperations.DMSRequestOperationsBase import DMSRequestOperationsBase

from DIRAC.MonitoringSystem.Client.MonitoringReporter import MonitoringReporter

########################################################################


class RemoveReplica(DMSRequestOperationsBase):
    """
    .. class:: RemoveReplica

    """

    def __init__(self, operation=None, csPath=None):
        """c'tor

        :param self: self reference
        :param Operation operation: operation to execute
        :param str csPath: CS path for this handler
        """
        # # base class ctor
        DMSRequestOperationsBase.__init__(self, operation, csPath)

    def __call__(self):
        """remove replicas"""

        # The flag  'rmsMonitoring' is set by the RequestTask and is False by default.
        # Here we use 'createRMSRecord' to create the ES record which is defined inside OperationHandlerBase.
        if self.rmsMonitoring:
            self.rmsMonitoringReporter = MonitoringReporter(monitoringType="RMSMonitoring")
        else:
            # # gMonitor stuff
            gMonitor.registerActivity(
                "RemoveReplicaAtt", "Replica removals attempted", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RemoveReplicaOK", "Successful replica removals", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )
            gMonitor.registerActivity(
                "RemoveReplicaFail", "Failed replica removals", "RequestExecutingAgent", "Files/min", gMonitor.OP_SUM
            )

        # # prepare list of targetSEs
        targetSEs = self.operation.targetSEList
        # # check targetSEs for removal
        bannedTargets = self.checkSEsRSS(targetSEs, access="RemoveAccess")
        if not bannedTargets["OK"]:
            if self.rmsMonitoring:
                for status in ["Attempted", "Failed"]:
                    self.rmsMonitoringReporter.addRecord(self.createRMSRecord(status, len(self.operation)))
                self.rmsMonitoringReporter.commit()
            else:
                gMonitor.addMark("RemoveReplicaAtt")
                gMonitor.addMark("RemoveReplicaFail")
            return bannedTargets

        if bannedTargets["Value"]:
            return S_OK("%s targets are banned for removal" % ",".join(bannedTargets["Value"]))

        # # get waiting files
        waitingFiles = self.getWaitingFilesList()
        # # and prepare dict
        toRemoveDict = dict((opFile.LFN, opFile) for opFile in waitingFiles)

        self.log.info("Todo: %s replicas to delete from %s SEs" % (len(toRemoveDict), len(targetSEs)))

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Attempted", len(toRemoveDict)))
        else:
            gMonitor.addMark("RemoveReplicaAtt", len(toRemoveDict) * len(targetSEs))

        # # keep status for each targetSE
        removalStatus = dict.fromkeys(toRemoveDict, None)
        for lfn in removalStatus:
            removalStatus[lfn] = dict.fromkeys(targetSEs, None)

        # # loop over targetSEs
        for targetSE in targetSEs:

            self.log.info("Removing replicas at %s" % targetSE)

            # # 1st step - bulk removal
            bulkRemoval = self._bulkRemoval(toRemoveDict, targetSE)
            if not bulkRemoval["OK"]:
                self.log.error("Bulk replica removal failed", bulkRemoval["Message"])

                if self.rmsMonitoring:
                    self.rmsMonitoringReporter.commit()

                return bulkRemoval

            # # report removal status for successful files
            if self.rmsMonitoring:
                self.rmsMonitoringReporter.addRecord(
                    self.createRMSRecord(
                        "Successful", len(([opFile for opFile in toRemoveDict.values() if not opFile.Error]))
                    )
                )
            else:
                gMonitor.addMark(
                    "RemoveReplicaOK", len([opFile for opFile in toRemoveDict.values() if not opFile.Error])
                )

            # # 2nd step - process the rest again
            toRetry = dict((lfn, opFile) for lfn, opFile in toRemoveDict.items() if opFile.Error)
            for lfn, opFile in toRetry.items():
                self._removeWithOwnerProxy(opFile, targetSE)
                if opFile.Error:
                    if self.rmsMonitoring:
                        self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Failed", 1))
                    else:
                        gMonitor.addMark("RemoveReplicaFail", 1)
                    removalStatus[lfn][targetSE] = opFile.Error
                else:
                    if self.rmsMonitoring:
                        self.rmsMonitoringReporter.addRecord(self.createRMSRecord("Successful", 1))
                    else:
                        gMonitor.addMark("RemoveReplicaOK", 1)

        # # update file status for waiting files
        failed = 0
        for opFile in self.operation:
            if opFile.Status == "Waiting":
                errors = list(set(error for error in removalStatus[opFile.LFN].values() if error))
                if errors:
                    opFile.Error = "\n".join(errors)
                    # This seems to be the only unrecoverable error
                    if "Write access not permitted for this credential" in opFile.Error:
                        failed += 1
                        opFile.Status = "Failed"
                else:
                    opFile.Status = "Done"

        if failed:
            self.operation.Error = "failed to remove %s replicas" % failed

        if self.rmsMonitoring:
            self.rmsMonitoringReporter.commit()

        return S_OK()

    def _bulkRemoval(self, toRemoveDict, targetSE):
        """remove replicas :toRemoveDict: at :targetSE:

        :param dict toRemoveDict: { lfn: opFile, ... }
        :param str targetSE: target SE name
        """
        # Clear the error
        for opFile in toRemoveDict.values():
            opFile.Error = ""
        removeReplicas = self.dm.removeReplica(targetSE, list(toRemoveDict))
        if not removeReplicas["OK"]:
            for opFile in toRemoveDict.values():
                opFile.Error = removeReplicas["Message"]
            return removeReplicas
        removeReplicas = removeReplicas["Value"]
        # # filter out failed
        for lfn, opFile in toRemoveDict.items():
            if lfn in removeReplicas["Failed"]:
                errorReason = str(removeReplicas["Failed"][lfn])
                # If the reason is that the file does not exist,
                # we consider the removal successful
                # TODO: use cmpError once the FC returns the proper error msg corresponding to ENOENT
                if "No such file" not in errorReason:
                    opFile.Error = errorReason
                    self.log.error("Failed removing lfn", "%s:%s" % (lfn, opFile.Error))

        return S_OK()

    def _removeWithOwnerProxy(self, opFile, targetSE):
        """remove opFile replica from targetSE using owner proxy

        :param File opFile: File instance
        :param str targetSE: target SE name
        """
        if "Write access not permitted for this credential" in opFile.Error:
            proxyFile = None
            if "DataManager" in self.shifter:
                # #  you're a data manager - save current proxy and get a new one for LFN and retry
                saveProxy = os.environ["X509_USER_PROXY"]
                try:
                    fileProxy = self.getProxyForLFN(opFile.LFN)
                    if not fileProxy["OK"]:
                        opFile.Error = fileProxy["Message"]
                    else:
                        proxyFile = fileProxy["Value"]
                        removeReplica = self.dm.removeReplica(targetSE, opFile.LFN)
                        if not removeReplica["OK"]:
                            opFile.Error = removeReplica["Message"]
                        else:
                            # Set or reset the error if all OK
                            opFile.Error = removeReplica["Value"]["Failed"].get(opFile.LFN, "")
                finally:
                    if proxyFile:
                        os.unlink(proxyFile)
                    # # put back request owner proxy to env
                    os.environ["X509_USER_PROXY"] = saveProxy



"""DIRAC Administrator API Class

All administrative functionality is exposed through the DIRAC Admin API.  Examples include
site banning and unbanning, WMS proxy uploading etc.

"""

from __future__ import print_function
from __future__ import absolute_import
from __future__ import division

__RCSID__ = "$Id$"

import six
import os

from DIRAC import gLogger, gConfig, S_OK, S_ERROR
from DIRAC.Core.Utilities.PromptUser import promptUser
from DIRAC.Core.Base.API import API
from DIRAC.ConfigurationSystem.Client.CSAPI import CSAPI
from DIRAC.ConfigurationSystem.Client.Helpers.Registry import getVOForGroup
from DIRAC.Core.Security.ProxyInfo import getProxyInfo
from DIRAC.FrameworkSystem.Client.ProxyManagerClient import gProxyManager
from DIRAC.FrameworkSystem.Client.NotificationClient import NotificationClient
from DIRAC.ResourceStatusSystem.Client.ResourceStatusClient import ResourceStatusClient
from DIRAC.ResourceStatusSystem.Client.ResourceStatus import ResourceStatus
from DIRAC.ResourceStatusSystem.Client.SiteStatus import SiteStatus
from DIRAC.WorkloadManagementSystem.Client.JobManagerClient import JobManagerClient
from DIRAC.WorkloadManagementSystem.Client.WMSAdministratorClient import WMSAdministratorClient
from DIRAC.WorkloadManagementSystem.Client.PilotManagerClient import PilotManagerClient

voName = ""
ret = getProxyInfo(disableVOMS=True)
if ret["OK"] and "group" in ret["Value"]:
    voName = getVOForGroup(ret["Value"]["group"])

COMPONENT_NAME = "/Interfaces/API/DiracAdmin"


class DiracAdmin(API):
    """Administrative functionalities"""

    #############################################################################
    def __init__(self):
        """Internal initialization of the DIRAC Admin API."""
        super(DiracAdmin, self).__init__()

        self.csAPI = CSAPI()

        self.dbg = False
        if gConfig.getValue(self.section + "/LogLevel", "DEBUG") == "DEBUG":
            self.dbg = True

        self.scratchDir = gConfig.getValue(self.section + "/ScratchDir", "/tmp")
        self.currentDir = os.getcwd()
        self.rssFlag = ResourceStatus().rssFlag
        self.sitestatus = SiteStatus()

    #############################################################################
    def uploadProxy(self):
        """Upload a proxy to the DIRAC WMS.  This method

        Example usage:

          >>> print diracAdmin.uploadProxy('dteam_pilot')
          {'OK': True, 'Value': 0L}

        :return: S_OK,S_ERROR

        :param permanent: Indefinitely update proxy
        :type permanent: boolean

        """
        return gProxyManager.uploadProxy()

    #############################################################################
    def setProxyPersistency(self, userDN, userGroup, persistent=True):
        """Set the persistence of a proxy in the Proxy Manager

        Example usage:

          >>> gLogger.notice(diracAdmin.setProxyPersistency( 'some DN', 'dirac group', True ))
          {'OK': True }

        :param userDN: User DN
        :type userDN: string
        :param userGroup: DIRAC Group
        :type userGroup: string
        :param persistent: Persistent flag
        :type persistent: boolean
        :return: S_OK,S_ERROR
        """
        return gProxyManager.setPersistency(userDN, userGroup, persistent)

    #############################################################################
    def checkProxyUploaded(self, userDN, userGroup, requiredTime):
        """Set the persistence of a proxy in the Proxy Manager

        Example usage:

          >>> gLogger.notice(diracAdmin.setProxyPersistency( 'some DN', 'dirac group', True ))
          {'OK': True, 'Value' : True/False }

        :param userDN: User DN
        :type userDN: string
        :param userGroup: DIRAC Group
        :type userGroup: string
        :param requiredTime: Required life time of the uploaded proxy
        :type requiredTime: boolean
        :return: S_OK,S_ERROR
        """
        return gProxyManager.userHasProxy(userDN, userGroup, requiredTime)

    #############################################################################
    def getSiteMask(self, printOutput=False, status="Active"):
        """Retrieve current site mask from WMS Administrator service.

        Example usage:

          >>> gLogger.notice(diracAdmin.getSiteMask())
          {'OK': True, 'Value': 0L}

        :return: S_OK,S_ERROR

        """

        result = self.sitestatus.getSites(siteState=status)
        if result["OK"]:
            sites = result["Value"]
            if printOutput:
                sites.sort()
                for site in sites:
                    gLogger.notice(site)

        return result

    #############################################################################
    def getBannedSites(self, printOutput=False):
        """Retrieve current list of banned  and probing sites.

        Example usage:

          >>> gLogger.notice(diracAdmin.getBannedSites())
          {'OK': True, 'Value': []}

        :return: S_OK,S_ERROR

        """

        bannedSites = self.sitestatus.getSites(siteState="Banned")
        if not bannedSites["OK"]:
            return bannedSites

        probingSites = self.sitestatus.getSites(siteState="Probing")
        if not probingSites["OK"]:
            return probingSites

        mergedList = sorted(bannedSites["Value"] + probingSites["Value"])

        if printOutput:
            gLogger.notice("\n".join(mergedList))

        return S_OK(mergedList)

    #############################################################################
    def getSiteSection(self, site, printOutput=False):
        """Simple utility to get the list of CEs for DIRAC site name.

        Example usage:

          >>> gLogger.notice(diracAdmin.getSiteSection('LCG.CERN.ch'))
          {'OK': True, 'Value':}

        :return: S_OK,S_ERROR
        """
        gridType = site.split(".")[0]
        if not gConfig.getSections("/Resources/Sites/%s" % (gridType))["OK"]:
            return S_ERROR("/Resources/Sites/%s is not a valid site section" % (gridType))

        result = gConfig.getOptionsDict("/Resources/Sites/%s/%s" % (gridType, site))
        if printOutput and result["OK"]:
            gLogger.notice(self.pPrint.pformat(result["Value"]))
        return result

    #############################################################################
    def allowSite(self, site, comment, printOutput=False):
        """Adds the site to the site mask.

        Example usage:

          >>> gLogger.notice(diracAdmin.allowSite())
          {'OK': True, 'Value': }

        :return: S_OK,S_ERROR

        """
        result = self._checkSiteIsValid(site)
        if not result["OK"]:
            return result

        result = self.getSiteMask(status="Active")
        if not result["OK"]:
            return result
        siteMask = result["Value"]
        if site in siteMask:
            if printOutput:
                gLogger.notice("Site %s is already Active" % site)
            return S_OK("Site %s is already Active" % site)

        if self.rssFlag:
            result = self.sitestatus.setSiteStatus(site, "Active", comment)
        else:
            result = WMSAdministratorClient().allowSite(site, comment)
        if not result["OK"]:
            return result

        if printOutput:
            gLogger.notice("Site %s status is set to Active" % site)

        return result

    #############################################################################
    def getSiteMaskLogging(self, site=None, printOutput=False):
        """Retrieves site mask logging information.

        Example usage:

          >>> gLogger.notice(diracAdmin.getSiteMaskLogging('LCG.AUVER.fr'))
          {'OK': True, 'Value': }

        :return: S_OK,S_ERROR
        """
        result = self._checkSiteIsValid(site)
        if not result["OK"]:
            return result

        if self.rssFlag:
            result = ResourceStatusClient().selectStatusElement("Site", "History", name=site)
        else:
            result = WMSAdministratorClient().getSiteMaskLogging(site)

        if not result["OK"]:
            return result

        if printOutput:
            if site:
                gLogger.notice("\nSite Mask Logging Info for %s\n" % site)
            else:
                gLogger.notice("\nAll Site Mask Logging Info\n")

            sitesLogging = result["Value"]
            if isinstance(sitesLogging, dict):
                for siteName, tupleList in sitesLogging.items():  # can be an iterator
                    if not siteName:
                        gLogger.notice("\n===> %s\n" % siteName)
                    for tup in tupleList:
                        stup = str(tup[0]).ljust(8) + str(tup[1]).ljust(20)
                        stup += "( " + str(tup[2]).ljust(len(str(tup[2]))) + ' )  "' + str(tup[3]) + '"'
                        gLogger.notice(stup)
                    gLogger.notice(" ")
            elif isinstance(sitesLogging, list):
                sitesLoggingList = [(sl[1], sl[3], sl[4]) for sl in sitesLogging]
                for siteLog in sitesLoggingList:
                    gLogger.notice(siteLog)

        return S_OK()

    #############################################################################
    def banSite(self, site, comment, printOutput=False):
        """Removes the site from the site mask.

        Example usage:

          >>> gLogger.notice(diracAdmin.banSite())
          {'OK': True, 'Value': }

        :return: S_OK,S_ERROR

        """
        result = self._checkSiteIsValid(site)
        if not result["OK"]:
            return result

        mask = self.getSiteMask(status="Banned")
        if not mask["OK"]:
            return mask
        siteMask = mask["Value"]
        if site in siteMask:
            if printOutput:
                gLogger.notice("Site %s is already Banned" % site)
            return S_OK("Site %s is already Banned" % site)

        if self.rssFlag:
            result = self.sitestatus.setSiteStatus(site, "Banned", comment)
        else:
            result = WMSAdministratorClient().banSite(site, comment)
        if not result["OK"]:
            return result

        if printOutput:
            gLogger.notice("Site %s status is set to Banned" % site)

        return result

    #############################################################################
    def getServicePorts(self, setup="", printOutput=False):
        """Checks the service ports for the specified setup.  If not given this is
        taken from the current installation (/DIRAC/Setup)

        Example usage:

          >>> gLogger.notice(diracAdmin.getServicePorts())
          {'OK': True, 'Value':''}

        :return: S_OK,S_ERROR

        """
        if not setup:
            setup = gConfig.getValue("/DIRAC/Setup", "")

        setupList = gConfig.getSections("/DIRAC/Setups", [])
        if not setupList["OK"]:
            return S_ERROR("Could not get /DIRAC/Setups sections")
        setupList = setupList["Value"]
        if setup not in setupList:
            return S_ERROR("Setup %s is not in allowed list: %s" % (setup, ", ".join(setupList)))

        serviceSetups = gConfig.getOptionsDict("/DIRAC/Setups/%s" % setup)
        if not serviceSetups["OK"]:
            return S_ERROR("Could not get /DIRAC/Setups/%s options" % setup)
        serviceSetups = serviceSetups["Value"]  # dict
        systemList = gConfig.getSections("/Systems")
        if not systemList["OK"]:
            return S_ERROR("Could not get Systems sections")
        systemList = systemList["Value"]
        result = {}
        for system in systemList:
            if system in serviceSetups:
                path = "/Systems/%s/%s/Services" % (system, serviceSetups[system])
                servicesList = gConfig.getSections(path)
                if not servicesList["OK"]:
                    self.log.warn("Could not get sections in %s" % path)
                else:
                    servicesList = servicesList["Value"]
                    if not servicesList:
                        servicesList = []
                    self.log.verbose("System: %s ServicesList: %s" % (system, ", ".join(servicesList)))
                    for service in servicesList:
                        spath = "%s/%s/Port" % (path, service)
                        servicePort = gConfig.getValue(spath, 0)
                        if servicePort:
                            self.log.verbose("Found port for %s/%s = %s" % (system, service, servicePort))
                            result["%s/%s" % (system, service)] = servicePort
                        else:
                            self.log.warn("No port found for %s" % spath)
            else:
                self.log.warn("%s is not defined in /DIRAC/Setups/%s" % (system, setup))

        if printOutput:
            gLogger.notice(self.pPrint.pformat(result))

        return S_OK(result)

    #############################################################################
    def getProxy(self, userDN, userGroup, validity=43200, limited=False):
        """Retrieves a proxy with default 12hr validity and stores
        this in a file in the local directory by default.

        Example usage:

          >>> gLogger.notice(diracAdmin.getProxy())
          {'OK': True, 'Value': }

        :return: S_OK,S_ERROR

        """
        return gProxyManager.downloadProxy(userDN, userGroup, limited=limited, requiredTimeLeft=validity)

    #############################################################################
    def getVOMSProxy(self, userDN, userGroup, vomsAttr=False, validity=43200, limited=False):
        """Retrieves a proxy with default 12hr validity and VOMS extensions and stores
        this in a file in the local directory by default.

        Example usage:

          >>> gLogger.notice(diracAdmin.getVOMSProxy())
          {'OK': True, 'Value': }

        :return: S_OK,S_ERROR

        """
        return gProxyManager.downloadVOMSProxy(
            userDN, userGroup, limited=limited, requiredVOMSAttribute=vomsAttr, requiredTimeLeft=validity
        )

    #############################################################################
    def getPilotProxy(self, userDN, userGroup, validity=43200):
        """Retrieves a pilot proxy with default 12hr validity and stores
        this in a file in the local directory by default.

        Example usage:

          >>> gLogger.notice(diracAdmin.getVOMSProxy())
          {'OK': True, 'Value': }

        :return: S_OK,S_ERROR

        """

        return gProxyManager.getPilotProxyFromDIRACGroup(userDN, userGroup, requiredTimeLeft=validity)

    #############################################################################
    def resetJob(self, jobID):
        """Reset a job or list of jobs in the WMS.  This operation resets the reschedule
        counter for a job or list of jobs and allows them to run as new.

        Example::

          >>> gLogger.notice(dirac.reset(12345))
          {'OK': True, 'Value': [12345]}

        :param job: JobID
        :type job: integer or list of integers
        :return: S_OK,S_ERROR

        """
        if isinstance(jobID, six.string_types):
            try:
                jobID = int(jobID)
            except Exception as x:
                return self._errorReport(str(x), "Expected integer or convertible integer for existing jobID")
        elif isinstance(jobID, list):
            try:
                jobID = [int(job) for job in jobID]
            except Exception as x:
                return self._errorReport(str(x), "Expected integer or convertible integer for existing jobIDs")

        result = JobManagerClient(useCertificates=False).resetJob(jobID)
        return result

    #############################################################################
    def getJobPilotOutput(self, jobID, directory=""):
        """Retrieve the pilot output for an existing job in the WMS.
        The output will be retrieved in a local directory unless
        otherwise specified.

          >>> gLogger.notice(dirac.getJobPilotOutput(12345))
          {'OK': True, StdOut:'',StdError:''}

        :param job: JobID
        :type job: integer or string
        :return: S_OK,S_ERROR
        """
        if not directory:
            directory = self.currentDir

        if not os.path.exists(directory):
            return self._errorReport("Directory %s does not exist" % directory)

        result = WMSAdministratorClient().getJobPilotOutput(jobID)
        if not result["OK"]:
            return result

        outputPath = "%s/pilot_%s" % (directory, jobID)
        if os.path.exists(outputPath):
            self.log.info("Remove %s and retry to continue" % outputPath)
            return S_ERROR("Remove %s and retry to continue" % outputPath)

        if not os.path.exists(outputPath):
            self.log.verbose("Creating directory %s" % outputPath)
            os.mkdir(outputPath)

        outputs = result["Value"]
        if "StdOut" in outputs:
            stdout = "%s/std.out" % (outputPath)
            with open(stdout, "w") as fopen:
                fopen.write(outputs["StdOut"])
            self.log.verbose("Standard output written to %s" % (stdout))
        else:
            self.log.warn("No standard output returned")

        if "StdError" in outputs:
            stderr = "%s/std.err" % (outputPath)
            with open(stderr, "w") as fopen:
                fopen.write(outputs["StdError"])
            self.log.verbose("Standard error written to %s" % (stderr))
        else:
            self.log.warn("No standard error returned")

        self.log.always("Outputs retrieved in %s" % outputPath)
        return result

    #############################################################################
    def getPilotOutput(self, gridReference, directory=""):
        """Retrieve the pilot output  (std.out and std.err) for an existing job in the WMS.

          >>> gLogger.notice(dirac.getJobPilotOutput(12345))
          {'OK': True, 'Value': {}}

        :param job: JobID
        :type job: integer or string
        :return: S_OK,S_ERROR
        """
        if not isinstance(gridReference, six.string_types):
            return self._errorReport("Expected string for pilot reference")

        if not directory:
            directory = self.currentDir

        if not os.path.exists(directory):
            return self._errorReport("Directory %s does not exist" % directory)

        result = PilotManagerClient().getPilotOutput(gridReference)
        if not result["OK"]:
            return result

        gridReferenceSmall = gridReference.split("/")[-1]
        if not gridReferenceSmall:
            gridReferenceSmall = "reference"
        outputPath = "%s/pilot_%s" % (directory, gridReferenceSmall)

        if os.path.exists(outputPath):
            self.log.info("Remove %s and retry to continue" % outputPath)
            return S_ERROR("Remove %s and retry to continue" % outputPath)

        if not os.path.exists(outputPath):
            self.log.verbose("Creating directory %s" % outputPath)
            os.mkdir(outputPath)

        outputs = result["Value"]
        if "StdOut" in outputs:
            stdout = "%s/std.out" % (outputPath)
            with open(stdout, "w") as fopen:
                fopen.write(outputs["StdOut"])
            self.log.info("Standard output written to %s" % (stdout))
        else:
            self.log.warn("No standard output returned")

        if "StdErr" in outputs:
            stderr = "%s/std.err" % (outputPath)
            with open(stderr, "w") as fopen:
                fopen.write(outputs["StdErr"])
            self.log.info("Standard error written to %s" % (stderr))
        else:
            self.log.warn("No standard error returned")

        self.log.always("Outputs retrieved in %s" % outputPath)
        return result

    #############################################################################
    def getPilotInfo(self, gridReference):
        """Retrieve info relative to a pilot reference

          >>> gLogger.notice(dirac.getPilotInfo(12345))
          {'OK': True, 'Value': {}}

        :param gridReference: Pilot Job Reference
        :type gridReference: string
        :return: S_OK,S_ERROR
        """
        if not isinstance(gridReference, six.string_types):
            return self._errorReport("Expected string for pilot reference")

        result = PilotManagerClient().getPilotInfo(gridReference)
        return result

    #############################################################################
    def killPilot(self, gridReference):
        """Kill the pilot specified

          >>> gLogger.notice(dirac.getPilotInfo(12345))
          {'OK': True, 'Value': {}}

        :param gridReference: Pilot Job Reference
        :return: S_OK,S_ERROR
        """
        if not isinstance(gridReference, six.string_types):
            return self._errorReport("Expected string for pilot reference")

        result = PilotManagerClient().killPilot(gridReference)
        return result

    #############################################################################
    def getPilotLoggingInfo(self, gridReference):
        """Retrieve the pilot logging info for an existing job in the WMS.

          >>> gLogger.notice(dirac.getPilotLoggingInfo(12345))
          {'OK': True, 'Value': {"The output of the command"}}

        :param gridReference: Gridp pilot job reference Id
        :type gridReference: string
        :return: S_OK,S_ERROR
        """
        if not isinstance(gridReference, six.string_types):
            return self._errorReport("Expected string for pilot reference")

        return PilotManagerClient().getPilotLoggingInfo(gridReference)

    #############################################################################
    def getJobPilots(self, jobID):
        """Extract the list of submitted pilots and their status for a given
        jobID from the WMS.  Useful information is printed to the screen.

          >>> gLogger.notice(dirac.getJobPilots())
          {'OK': True, 'Value': {PilotID:{StatusDict}}}

        :param job: JobID
        :type job: integer or string
        :return: S_OK,S_ERROR

        """
        if isinstance(jobID, six.string_types):
            try:
                jobID = int(jobID)
            except Exception as x:
                return self._errorReport(str(x), "Expected integer or string for existing jobID")

        result = PilotManagerClient().getPilots(jobID)
        if result["OK"]:
            gLogger.notice(self.pPrint.pformat(result["Value"]))
        return result

    #############################################################################
    def getPilotSummary(self, startDate="", endDate=""):
        """Retrieve the pilot output for an existing job in the WMS.  Summary is
        printed at INFO level, full dictionary of results also returned.

          >>> gLogger.notice(dirac.getPilotSummary())
          {'OK': True, 'Value': {CE:{Status:Count}}}

        :param job: JobID
        :type job: integer or string
        :return: S_OK,S_ERROR
        """
        result = PilotManagerClient().getPilotSummary(startDate, endDate)
        if not result["OK"]:
            return result

        ceDict = result["Value"]
        headers = "CE".ljust(28)
        i = 0
        for ce, summary in ceDict.items():
            states = list(summary)
            if len(states) > i:
                i = len(states)

        for i in range(i):
            headers += "Status".ljust(12) + "Count".ljust(12)
        gLogger.notice(headers)

        for ce, summary in ceDict.items():
            line = ce.ljust(28)
            states = sorted(summary)
            for state in states:
                count = str(summary[state])
                line += state.ljust(12) + count.ljust(12)
            gLogger.notice(line)

        return result

    #############################################################################
    def setSiteProtocols(self, site, protocolsList, printOutput=False):
        """
        Allows to set the defined protocols for each SE for a given site.
        """
        result = self._checkSiteIsValid(site)
        if not result["OK"]:
            return result

        siteSection = "/Resources/Sites/%s/%s/SE" % (site.split(".")[0], site)
        siteSEs = gConfig.getValue(siteSection, [])
        if not siteSEs:
            return S_ERROR("No SEs found for site %s in section %s" % (site, siteSection))

        defaultProtocols = gConfig.getValue("/Resources/StorageElements/DefaultProtocols", [])
        self.log.verbose("Default list of protocols are", ", ".join(defaultProtocols))

        for protocol in protocolsList:
            if protocol not in defaultProtocols:
                return S_ERROR(
                    "Requested to set protocol %s in list but %s is not "
                    "in default list of protocols:\n%s" % (protocol, protocol, ", ".join(defaultProtocols))
                )

        modifiedCS = False
        result = promptUser(
            "Do you want to add the following default protocols:"
            " %s for SE(s):\n%s" % (", ".join(protocolsList), ", ".join(siteSEs))
        )
        if not result["OK"]:
            return result
        if result["Value"].lower() != "y":
            self.log.always("No protocols will be added")
            return S_OK()

        for se in siteSEs:
            sections = gConfig.getSections("/Resources/StorageElements/%s/" % (se))
            if not sections["OK"]:
                return sections
            for section in sections["Value"]:
                if gConfig.getValue("/Resources/StorageElements/%s/%s/ProtocolName" % (se, section), "") == "SRM2":
                    path = "/Resources/StorageElements/%s/%s/ProtocolsList" % (se, section)
                    self.log.verbose("Setting %s to %s" % (path, ", ".join(protocolsList)))
                    result = self.csSetOption(path, ", ".join(protocolsList))
                    if not result["OK"]:
                        return result
                    modifiedCS = True

        if modifiedCS:
            result = self.csCommitChanges(False)
            if not result["OK"]:
                return S_ERROR("CS Commit failed with message = %s" % (result["Message"]))
            else:
                if printOutput:
                    gLogger.notice("Successfully committed changes to CS")
        else:
            if printOutput:
                gLogger.notice("No modifications to CS required")

        return S_OK()

    #############################################################################
    def csSetOption(self, optionPath, optionValue):
        """
        Function to modify an existing value in the CS.
        """
        return self.csAPI.setOption(optionPath, optionValue)

    #############################################################################
    def csSetOptionComment(self, optionPath, comment):
        """
        Function to modify an existing value in the CS.
        """
        return self.csAPI.setOptionComment(optionPath, comment)

    #############################################################################
    def csModifyValue(self, optionPath, newValue):
        """
        Function to modify an existing value in the CS.
        """
        return self.csAPI.modifyValue(optionPath, newValue)

    #############################################################################
    def csRegisterUser(self, username, properties):
        """
        Registers a user in the CS.

            - username: Username of the user (easy;)
            - properties: Dict containing:
                - DN
                - groups : list/tuple of groups the user belongs to
                - <others> : More properties of the user, like mail

        """
        return self.csAPI.addUser(username, properties)

    #############################################################################
    def csDeleteUser(self, user):
        """
        Deletes a user from the CS. Can take a list of users
        """
        return self.csAPI.deleteUsers(user)

    #############################################################################
    def csModifyUser(self, username, properties, createIfNonExistant=False):
        """
        Modify a user in the CS. Takes the same params as in addUser and
        applies the changes
        """
        return self.csAPI.modifyUser(username, properties, createIfNonExistant)

    #############################################################################
    def csListUsers(self, group=False):
        """
        Lists the users in the CS. If no group is specified return all users.
        """
        return self.csAPI.listUsers(group)

    #############################################################################
    def csDescribeUsers(self, mask=False):
        """
        List users and their properties in the CS.
        If a mask is given, only users in the mask will be returned
        """
        return self.csAPI.describeUsers(mask)

    #############################################################################
    def csModifyGroup(self, groupname, properties, createIfNonExistant=False):
        """
        Modify a user in the CS. Takes the same params as in addGroup and applies
        the changes
        """
        return self.csAPI.modifyGroup(groupname, properties, createIfNonExistant)

    #############################################################################
    def csListHosts(self):
        """
        Lists the hosts in the CS
        """
        return self.csAPI.listHosts()

    #############################################################################
    def csDescribeHosts(self, mask=False):
        """
        Gets extended info for the hosts in the CS
        """
        return self.csAPI.describeHosts(mask)

    #############################################################################
    def csModifyHost(self, hostname, properties, createIfNonExistant=False):
        """
        Modify a host in the CS. Takes the same params as in addHost and applies
        the changes
        """
        return self.csAPI.modifyHost(hostname, properties, createIfNonExistant)

    #############################################################################
    def csListGroups(self):
        """
        Lists groups in the CS
        """
        return self.csAPI.listGroups()

    #############################################################################
    def csDescribeGroups(self, mask=False):
        """
        List groups and their properties in the CS.
        If a mask is given, only groups in the mask will be returned
        """
        return self.csAPI.describeGroups(mask)

    #############################################################################
    def csSyncUsersWithCFG(self, usersCFG):
        """
        Synchronize users in cfg with its contents
        """
        return self.csAPI.syncUsersWithCFG(usersCFG)

    #############################################################################
    def csCommitChanges(self, sortUsers=True):
        """
        Commit the changes in the CS
        """
        return self.csAPI.commitChanges(sortUsers=False)

    #############################################################################
    def sendMail(self, address, subject, body, fromAddress=None, localAttempt=True, html=False):
        """
        Send mail to specified address with body.
        """
        notification = NotificationClient()
        return notification.sendMail(address, subject, body, fromAddress, localAttempt, html)

    #############################################################################
    def sendSMS(self, userName, body, fromAddress=None):
        """
        Send mail to specified address with body.
        """
        if len(body) > 160:
            return S_ERROR("Exceeded maximum SMS length of 160 characters")
        notification = NotificationClient()
        return notification.sendSMS(userName, body, fromAddress)


# EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#EOF#



""" Executor Module

    Just provides a number of functions used by all executors
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from DIRAC import S_OK, S_ERROR, gConfig, gLogger, rootPath
from DIRAC.ConfigurationSystem.Client import PathFinder
from DIRAC.Core.Utilities.Shifter import setupShifterProxyInEnv
from DIRAC.Core.Utilities.ReturnValues import isReturnStructure

__RCSID__ = "$Id$"


class ExecutorModule(object):
    """All executors should inherit from this module"""

    @classmethod
    def _ex_initialize(cls, exeName, loadName):
        cls.__properties = {
            "fullName": exeName,
            "loadName": loadName,
            "section": PathFinder.getExecutorSection(exeName),
            "loadSection": PathFinder.getExecutorSection(loadName),
            "messagesProcessed": 0,
            "reconnects": 0,
            "setup": gConfig.getValue("/DIRAC/Setup", "Unknown"),
        }
        cls.__basePath = gConfig.getValue("/LocalSite/InstancePath", rootPath)
        cls.__defaults = {}
        cls.__defaults["MonitoringEnabled"] = True
        cls.__defaults["Enabled"] = True
        cls.__defaults["ControlDirectory"] = os.path.join(cls.__basePath, "control", *exeName.split("/"))
        cls.__defaults["WorkDirectory"] = os.path.join(cls.__basePath, "work", *exeName.split("/"))
        cls.__defaults["ReconnectRetries"] = 10
        cls.__defaults["ReconnectSleep"] = 5
        cls.__defaults["shifterProxy"] = ""
        cls.__defaults["shifterProxyLocation"] = os.path.join(cls.__defaults["WorkDirectory"], ".shifterCred")
        cls.__properties["shifterProxy"] = ""
        cls.__properties["shifterProxyLocation"] = os.path.join(cls.__defaults["WorkDirectory"], ".shifterCred")
        cls.__mindName = False
        cls.__mindExtraArgs = False
        cls.__freezeTime = 0
        cls.__fastTrackEnabled = True
        cls.log = gLogger.getSubLogger(exeName, child=False)

        try:
            result = cls.initialize()  # pylint: disable=no-member
        except Exception as excp:
            gLogger.exception("Exception while initializing %s" % loadName, lException=excp)
            return S_ERROR("Exception while initializing: %s" % str(excp))
        if not isReturnStructure(result):
            return S_ERROR("Executor %s does not return an S_OK/S_ERROR after initialization" % loadName)
        return result

    def __installShifterProxy(self):
        shifterProxy = self.ex_getProperty("shifterProxy")
        if not shifterProxy:
            return S_OK()
        location = "%s-%s" % (self.ex_getProperty("shifterProxyLocation"), shifterProxy)
        result = setupShifterProxyInEnv(shifterProxy, location)
        if not result["OK"]:
            self.log.error("Cannot set shifter proxy: %s" % result["Message"])
        return result

    @classmethod
    def ex_setOption(cls, optName, value):
        cls.__defaults[optName] = value

    @classmethod
    def ex_getOption(cls, optName, defaultValue=None):
        if defaultValue is None:
            if optName in cls.__defaults:
                defaultValue = cls.__defaults[optName]
        if optName and optName[0] == "/":
            return gConfig.getValue(optName, defaultValue)
        for section in (cls.__properties["section"], cls.__properties["loadSection"]):
            result = gConfig.getOption("%s/%s" % (section, optName), defaultValue)
            if result["OK"]:
                return result["Value"]
        return defaultValue

    @classmethod
    def ex_setProperty(cls, optName, value):
        cls.__properties[optName] = value

    @classmethod
    def ex_getProperty(cls, optName):
        return cls.__properties[optName]

    @classmethod
    def ex_enabled(cls):
        return cls.ex_getOption("Enabled")

    @classmethod
    def ex_setMind(cls, mindName, **extraArgs):
        cls.__mindName = mindName
        cls.__mindExtraArgs = extraArgs

    @classmethod
    def ex_getMind(cls):
        return cls.__mindName

    @classmethod
    def ex_getExtraArguments(cls):
        return cls.__mindExtraArgs

    def __serialize(self, taskId, taskObj):
        try:
            result = self.serializeTask(taskObj)
        except Exception as excp:
            gLogger.exception("Exception while serializing task %s" % taskId, lException=excp)
            return S_ERROR("Cannot serialize task %s: %s" % (taskId, str(excp)))
        if not isReturnStructure(result):
            raise Exception("serializeTask does not return a return structure")
        return result

    def __deserialize(self, taskId, taskStub):
        try:
            result = self.deserializeTask(taskStub)
        except Exception as excp:
            gLogger.exception("Exception while deserializing task %s" % taskId, lException=excp)
            return S_ERROR("Cannot deserialize task %s: %s" % (taskId, str(excp)))
        if not isReturnStructure(result):
            raise Exception("deserializeTask does not return a return structure")
        return result

    def _ex_processTask(self, taskId, taskStub):
        self.__properties["shifterProxy"] = self.ex_getOption("shifterProxy")
        self.__freezeTime = 0
        self.__fastTrackEnabled = True
        self.log.verbose("Task %s: Received" % str(taskId))
        result = self.__deserialize(taskId, taskStub)
        if not result["OK"]:
            self.log.error("Can not deserialize task", "Task %s: %s" % (str(taskId), result["Message"]))
            return result
        taskObj = result["Value"]
        # Shifter proxy?
        result = self.__installShifterProxy()
        if not result["OK"]:
            return result
        # Execute!
        result = self.processTask(taskId, taskObj)
        if not isReturnStructure(result):
            raise Exception("processTask does not return a return structure")
        if not result["OK"]:
            return result
        # If there's a result, serialize it again!
        if result["Value"]:
            taskObj = result["Value"]
        # Serialize again
        result = self.__serialize(taskId, taskObj)
        if not result["OK"]:
            self.log.verbose("Task %s: Cannot serialize: %s" % (str(taskId), result["Message"]))
            return result
        taskStub = result["Value"]
        # Try fast track
        fastTrackType = False
        if not self.__freezeTime and self.__fastTrackEnabled:
            result = self.fastTrackDispatch(taskId, taskObj)
            if not result["OK"]:
                self.log.error("FastTrackDispatch failed for job", "%s: %s" % (taskId, result["Message"]))
            else:
                fastTrackType = result["Value"]

        # EOP
        return S_OK((taskStub, self.__freezeTime, fastTrackType))

    ####
    # Callable functions
    ####

    def freezeTask(self, freezeTime):
        self.__freezeTime = freezeTime

    def isTaskFrozen(self):
        return self.__freezeTime

    def disableFastTrackForTask(self):
        self.__fastTrackEnabled = True

    ###
    #  Fast-track tasks
    ###

    def fastTrackDispatch(self, taskId, taskObj):
        return S_OK()

    ####
    # Need to overwrite this functions
    ####

    def serializeTask(self, taskObj):
        raise Exception("Method serializeTask has to be coded!")

    def deserializeTask(self, taskStub):
        raise Exception("Method deserializeTask has to be coded!")

    def processTask(self, taskId, taskObj):
        raise Exception("Method processTask has to be coded!")



"""
The Monitoring/Monitoring service interacts with the ElasticSearch backend
exposed by MonitoringDB.

.. literalinclude:: ../ConfigTemplate.cfg
  :start-after: ##BEGIN Monitoring
  :end-before: ##END
  :dedent: 2
  :caption: Monitoring options


"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import datetime
import os
import six

from DIRAC import gLogger, S_OK, S_ERROR, gConfig
from DIRAC.Core.DISET.RequestHandler import RequestHandler
from DIRAC.Core.Utilities import Time
from DIRAC.Core.Utilities.ObjectLoader import ObjectLoader
from DIRAC.Core.Utilities.Plotting import gDataCache
from DIRAC.Core.Utilities.Plotting.FileCoding import extractRequestFromFileId
from DIRAC.Core.Utilities.Plotting.Plots import generateErrorMessagePlot
from DIRAC.Core.Utilities.File import mkDir

from DIRAC.MonitoringSystem.private.MainReporter import MainReporter

__RCSID__ = "$Id$"


class MonitoringHandler(RequestHandler):

    """
    .. class:: MonitoringHandler

    :param dict __reportRequestDict: contains the arguments used to create a certain plot
    :param object __db: used to retrieve the data from the db.

    """

    __reportRequestDict = {
        "typeName": six.string_types,
        "reportName": six.string_types,
        "startTime": Time._allDateTypes,
        "endTime": Time._allDateTypes,
        "condDict": dict,
        "grouping": six.string_types,
        "extraArgs": dict,
    }

    __db = None

    @classmethod
    def initializeHandler(cls, serviceInfo):
        try:
            result = ObjectLoader().loadObject("MonitoringSystem.DB.MonitoringDB", "MonitoringDB")
            if not result["OK"]:
                return result
            cls.__db = result["Value"]()
        except RuntimeError as excp:
            return S_ERROR("Can't connect to DB: %s" % excp)

        reportSection = serviceInfo["serviceSectionPath"]
        dataPath = gConfig.getValue("%s/DataLocation" % reportSection, "data/monitoringPlots")
        gLogger.info("Data will be written into %s" % dataPath)
        mkDir(dataPath)
        try:
            testFile = "%s/moni.plot.test" % dataPath
            with open(testFile, "w") as _:
                os.unlink(testFile)
        except IOError as err:
            gLogger.fatal("Can't write to %s" % dataPath, err)
            return S_ERROR("Data location is not writable: %s" % repr(err))
        gDataCache.setGraphsLocation(dataPath)

        return S_OK()

    types_listUniqueKeyValues = [six.string_types]

    def export_listUniqueKeyValues(self, typeName):
        """
        :param str typeName: is the monitoring type registered in the Types.

        :return: S_OK({key:[]}) or S_ERROR()   The key is element of the __keyFields of the BaseType
        """
        setup = self.serviceInfoDict.get("clientSetup", None)
        if not setup:
            return S_ERROR("FATAL ERROR:  Problem with the service configuration!")
        # NOTE: we can apply some policies if it will be needed!
        return self.__db.getKeyValues(typeName)

    types_listReports = [six.string_types]

    def export_listReports(self, typeName):
        """
        :param str typeName: monitoring type for example WMSHistory

        :return: S_OK([]) or S_ERROR() the list of available plots
        """

        reporter = MainReporter(self.__db, self.serviceInfoDict["clientSetup"])
        return reporter.list(typeName)

    def transfer_toClient(self, fileId, token, fileHelper):
        """
        Get graphs data

        :param str fileId: encoded plot attributes
        :param object token: ???
        :param DIRAC.Core.DISET.private.FileHelper.FileHelper fileHelper:

        """

        # First check if we've got to generate the plot
        if len(fileId) > 5 and fileId[1] == ":":
            gLogger.info("Seems the file request is a plot generation request!")
            try:
                result = self.__generatePlotFromFileId(fileId)
            except Exception as e:  # pylint: disable=broad-except
                gLogger.exception("Exception while generating plot", str(e))
                result = S_ERROR("Error while generating plot: %s" % str(e))
            if not result["OK"]:
                self.__sendErrorAsImg(result["Message"], fileHelper)
                fileHelper.sendEOF()
                return result
            fileId = result["Value"]

        retVal = gDataCache.getPlotData(fileId)
        if not retVal["OK"]:
            self.__sendErrorAsImg(retVal["Message"], fileHelper)
            return retVal
        retVal = fileHelper.sendData(retVal["Value"])
        if not retVal["OK"]:
            return retVal
        fileHelper.sendEOF()
        return S_OK()

    def __generatePlotFromFileId(self, fileId):
        """
        It create the plots using the encode parameters

        :param str fileId: the encoded plot attributes
        :return S_OK or S_ERROR returns the file name
        """

        result = extractRequestFromFileId(fileId)
        if not result["OK"]:
            return result
        plotRequest = result["Value"]
        gLogger.info("Generating the plots..")
        result = self.export_generatePlot(plotRequest)
        if not result["OK"]:
            gLogger.error("Error while generating the plots", result["Message"])
            return result
        fileToReturn = "plot"
        if "extraArgs" in plotRequest:
            extraArgs = plotRequest["extraArgs"]
            if "thumbnail" in extraArgs and extraArgs["thumbnail"]:
                fileToReturn = "thumbnail"
        gLogger.info("Returning %s file: %s " % (fileToReturn, result["Value"][fileToReturn]))
        return S_OK(result["Value"][fileToReturn])

    def __sendErrorAsImg(self, msgText, fileHelper):
        """
        In case of an error message a whcite plot is created with the error message.
        """

        retVal = generateErrorMessagePlot(msgText)
        if not retVal:
            retVal = fileHelper.sendData(retVal["Message"])
        else:
            retVal = fileHelper.sendData(retVal["Value"])
            if not retVal["OK"]:
                return retVal
            fileHelper.sendEOF()
        return S_OK()

    def __checkPlotRequest(self, reportRequest):
        """
        It check the plot attributes. We have to make sure that all attributes which are needed are provided.

        :param dict reportRequest: contains the plot attributes.
        """
        # If extraArgs is not there add it
        if "extraArgs" not in reportRequest:
            reportRequest["extraArgs"] = {}
        if not isinstance(reportRequest["extraArgs"], self.__reportRequestDict["extraArgs"]):
            return S_ERROR("Extra args has to be of type %s" % self.__reportRequestDict["extraArgs"])
        reportRequestExtra = reportRequest["extraArgs"]

        # Check sliding plots
        if "lastSeconds" in reportRequestExtra:
            try:
                lastSeconds = int(reportRequestExtra["lastSeconds"])
            except ValueError:
                gLogger.error("lastSeconds key must be a number")
                return S_ERROR("Value Error")
            if lastSeconds < 3600:
                return S_ERROR("lastSeconds must be more than 3600")
            now = Time.dateTime()  # this is an UTC time
            reportRequest["endTime"] = now
            reportRequest["startTime"] = now - datetime.timedelta(seconds=lastSeconds)
        else:
            # if end date is not there, just set it to now
            if not reportRequest.get("endTime"):
                # check the existence of the endTime it can be present and empty
                reportRequest["endTime"] = Time.dateTime()
        # Check keys
        for key in self.__reportRequestDict:
            if key not in reportRequest:
                return S_ERROR("Missing mandatory field %s in plot request" % key)

            if not isinstance(reportRequest[key], self.__reportRequestDict[key]):
                return S_ERROR(
                    "Type mismatch for field %s (%s), required one of %s"
                    % (key, str(type(reportRequest[key])), str(self.__reportRequestDict[key]))
                )
            if key in ("startTime", "endTime"):
                reportRequest[key] = int(Time.toEpoch(reportRequest[key]))

        return S_OK(reportRequest)

    types_generatePlot = [dict]

    def export_generatePlot(self, reportRequest):
        """
        It creates a plots for a given request

        :param dict reportRequest: contains the plot arguments...
        """
        retVal = self.__checkPlotRequest(reportRequest)
        if not retVal["OK"]:
            return retVal
        reporter = MainReporter(self.__db, self.serviceInfoDict["clientSetup"])
        reportRequest["generatePlot"] = True
        return reporter.generate(reportRequest, self.getRemoteCredentials())

    types_getReport = [dict]

    def export_getReport(self, reportRequest):
        """
        It is used to get the raw data used to create a plot.
        The reportRequest has the following keys:

        str typeName: the type of the monitoring
        str reportName: the name of the plotter used to create the plot for example: NumberOfJobs
        int startTime: epoch time, start time of the plot
        int endTime: epoch time, end time of the plot
        dict condDict: is the conditions used to gnerate the plot: {'Status':['Running'],'grouping': ['Site'] }
        str grouping: is the grouping of the data for example: 'Site'
        dict extraArgs: epoch time which can be last day, last week, last month

        :returns: S_OK or S_ERROR S_OK value is a dictionary which contains all values used to create the plot
        """
        retVal = self.__checkPlotRequest(reportRequest)
        if not retVal["OK"]:
            return retVal
        reporter = MainReporter(self.__db, self.serviceInfoDict["clientSetup"])
        reportRequest["generatePlot"] = False
        return reporter.generate(reportRequest, self.getRemoteCredentials())

    types_addMonitoringRecords = [six.string_types, list]

    def export_addMonitoringRecords(self, monitoringtype, data):
        """
        Bulk insert data directly to the given monitoring type.

        :param str monitoringtype: monitoring type name
        :param list data: list of documents
        :returns: S_OK or S_ERROR
        """

        retVal = self.__db.getIndexName(monitoringtype)
        if not retVal["OK"]:
            return retVal
        prefix = retVal["Value"]
        gLogger.debug("addMonitoringRecords:", prefix)
        return self.__db.bulk_index(prefix, data)

    types_addRecords = [six.string_types, six.string_types, list]

    def export_addRecords(self, indexname, monitoringType, data):
        """
        It is used to insert data directly to the database... The data will be inserted to the given index.

        :param str indexname: name of the index
        :param str monitoringType: type of the monitoring
        :param list data: data to insert
        :returns: S_OK or S_ERROR
        """
        setup = self.serviceInfoDict.get("clientSetup", "")
        indexname = "%s_%s" % (setup.lower(), indexname)
        gLogger.debug("Bulk index:", indexname)
        mapping = self.__db.getMapping(monitoringType)
        gLogger.debug("Mapping:", mapping)
        return self.__db.bulk_index(indexname, data, mapping)

    types_deleteIndex = [six.string_types]

    def export_deleteIndex(self, indexName):
        """
        It is used to delete an index!
        Note this is for experienced users!!!

        :param str indexName: name of the index
        """
        setup = self.serviceInfoDict.get("clientSetup", "")
        indexName = "%s_%s" % (setup.lower(), indexName)
        gLogger.debug("delete index:", indexName)
        return self.__db.deleteIndex(indexName)

    types_getLastDayData = [six.string_types, dict]

    def export_getLastDayData(self, typeName, condDict):
        """
        It returns the data from the last day index. Note: we create daily indexes.

        :param str typeName: name of the monitoring type
        :param dict condDict: conditions for the query

                       * key -> name of the field
                       * value -> list of possible values
        """

        return self.__db.getLastDayData(typeName, condDict)

    types_getLimitedDat = [six.string_types, dict, int]

    def export_getLimitedData(self, typeName, condDict, size):
        """
        Returns a list of records for a given selection.

        :param str typeName: name of the monitoring type
        :param dict condDict: conditions for the query

                       * key -> name of the field
                       * value -> list of possible values

        :param int size: Indicates how many entries should be retrieved from the log
        :return: Up to size entries for the given component from the database
        """
        return self.__db.getLimitedData(typeName, condDict, size)

    types_getDataForAGivenPeriod = [six.string_types, dict, six.string_types, six.string_types]

    def export_getDataForAGivenPeriod(self, typeName, condDict, initialDate="", endDate=""):
        """
        Retrieves the history of logging entries for the given component during a given given time period

        :param str typeName: name of the monitoring type
        :param dict condDict: conditions for the query

                       * key -> name of the field
                       * value -> list of possible values

        :param str initialDate: Indicates the start of the time period in the format 'DD/MM/YYYY hh:mm'
        :param str endDate: Indicate the end of the time period in the format 'DD/MM/YYYY hh:mm'
        :return: Entries from the database for the given component recorded between the initial and the end dates

        """
        return self.__db.getDataForAGivenPeriod(typeName, condDict, initialDate, endDate)

    types_put = [list, six.string_types]

    def export_put(self, recordsToInsert, monitoringType):
        """
        It is used to insert records to the db.

        :param recordsToInsert: records to be inserted to the db
        :param str monitoringType: monitoring type...
        :type recordsToInsert: python:list
        """

        return self.__db.put(recordsToInsert, monitoringType)

    types_pingDB = []

    def export_pingDB(self):
        """
        We can check, if the db is available.
        """
        return self.__db.pingDB()
